--------------------------------------------------------------------------------
-- Author:  Flavio Tanese
-- Politecnico di Torino 2018
--------------------------------------------------------------------------------
-- Multiply two sparse circulant binary matrices stored in a memory as the
-- positions of set bits in their first row, and return a "tentative" result
-- (which is unordered and not simplified modulo 2) in another location in
-- memory.
-- Controlling circuitry should take care of keeping all inputs in the
-- "function arguments" section constant until the module reported back,
-- and of ensuring validity of such inputs (no overlapping memory ranges and
-- so on).
-- The "start_i" signal should be kept high until "mult_done_o" goes high,
-- implementing a rudimentary handshake, but this is not strictly required if
-- the control circuitry operates on the same clock.

library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.system_params.all;
    use work.matrix_types.all;
    use work.matrix_mult_functions.all;

entity circulant_multiplication is
    port(
        -- control signals
        clk_i:          in  std_logic;
        rst_n_i:        in  std_logic;
        start_i:        in  std_logic;
        -- function arguments in (not latched)
        a_limit_i:      in  natural;    -- number of elements in 1st matrix
        a_base_i:       in  addr;       -- base address of 1st matrix
        b_limit_i:      in  natural;    -- number of elements in 2nd matrix
        b_base_i:       in  addr;       -- base address of 2nd matrix
        z_base_i:       in  addr;       -- base address of result matrix
        -- data, addresses and controls to memory
        a_i:            in  pos;        -- operand from 1st matrix
        a_addr_o:       out addr;       -- address for 1st matrix
        b_i:            in  pos;        -- operand from 2nd matrix
        b_addr_o:       out addr;       -- address for 2nd matrix
        z_o:            out pos;        -- Ltr result
        z_addr_o:       out addr;       -- address for Ltr result
        wr_o:           out std_logic;  -- write enable for memory
        -- report back once done
        z_limit_o:      out natural;    -- number of elements in result matrix
        mult_done_o:    out std_logic   -- high when done
    );
end entity circulant_multiplication;

architecture rtl of circulant_multiplication is

    -- assume inputs are kept constant by higher level state machine so we do
    -- not need to sample them on start

    signal      a_index:            natural;
    signal      b_index:            natural;
    signal      z_index:            natural;
    signal      next_a_index:       natural;
    signal      next_b_index:       natural;
    signal      next_z_index:       natural;

    type state_t is (IDLE, BUSY, DONE);
    signal      state:              state_t;
    signal      next_state:         state_t;

begin

    z_o         <=  a_i + b_i when a_i + b_i < P else
                    (a_i + b_i) mod P;

    a_addr_o    <=  a_base_i + a_index;
    b_addr_o    <=  b_base_i + b_index;
    z_addr_o    <=  z_base_i + z_index;

    -- count number of elements of result matrix, this can be done with a
    -- multiplier but we are not in a hurry and do not want a big footprint.
    -- Using z_index we get that for free!
    z_limit_o   <=  z_index;

    state_comb: process(
        state, start_i, a_limit_i, b_limit_i, next_a_index, next_b_index
    )
    begin
        case state is
            when IDLE =>
                if start_i = '1' then
                    next_state  <=  BUSY;
                else
                    next_state  <=  IDLE;
                end if;
            when BUSY =>
                -- exit this state only once all combinations have been done
                if b_index /= b_limit_i or a_index /= a_limit_i then
                    next_state  <=  BUSY;
                else
                    next_state  <=  DONE;
                end if;
            when DONE =>
                if start_i = '0' then
                    next_state  <=  IDLE;
                else
                    next_state  <=  DONE;
                end if;
            when others =>
                next_state  <=  IDLE;
        end case;
    end process state_comb;

    state_seq: process(rst_n_i, clk_i)
    begin
        if rst_n_i = '0' then
            state   <=  IDLE;
        elsif rising_edge(clk_i) then
            state   <=  next_state;
        end if;
    end process state_seq;

    output_comb: process(
        state, a_index, b_index, z_index
    )
    begin
        case state is
            when IDLE =>
                next_a_index    <=  0;
                next_b_index    <=  0;
                next_z_index    <=  0;
                wr_o            <=  '0';
                mult_done_o     <=  '0';
            when BUSY =>
                if a_index < a_limit_i then
                    next_a_index    <=  a_index + 1;
                    next_b_index    <=  b_index;
                else
                    next_a_index    <=  0;
                    next_b_index    <=  b_index + 1;
                end if;
                next_z_index    <=  z_index + 1;
                wr_o            <=  '1';
                mult_done_o     <=  '0';
            when DONE =>
                next_a_index    <=  a_index;
                next_b_index    <=  b_index;
                next_z_index    <=  z_index;
                wr_o            <=  '0';
                mult_done_o     <=  '1';
            when others =>  -- behave like IDLE
                next_a_index    <=  0;
                next_b_index    <=  0;
                next_z_index    <=  0;
                wr_o            <=  '0';
                mult_done_o     <=  '0';
        end case;
    end process output_comb;

    output_seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            a_index     <=  0;
            b_index     <=  0;
            z_index     <=  0;
        elsif rising_edge(clk_i) then
            a_index     <=  next_a_index;
            b_index     <=  next_b_index;
            z_index     <=  next_z_index;
        end if;
    end process output_seq;

end architecture rtl;
