library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.system_params.all;
    use work.matrix_types.all;
    use work.matrix_mult_functions.all;

entity circulant_sum is
    port(
        -- control signals
        clk_i:      in  std_logic;
        rst_n_i:    in  std_logic;
        start_i:    in  std_logic;
        -- function arguments in (not latched)
        a_limit_i:  in  natural;    -- number of elements in 1st matrix
        a_base_i:   in  addr;       -- base address of 1st matrix
        b_limit_i:  in  natural;    -- number of elements in 2nd matrix
        b_base_i:   in  addr;       -- base address of 2nd matrix
        z_base_i:   in  addr;       -- base address of result matrix
        -- data, addresses and controls to memory
        a_i:        in  pos;
        a_addr_o:   out addr;
        b_i:        in  pos;
        b_addr_o:   out addr;
        z_o:        out pos;
        z_addr_o:   out addr;
        wr_o:       out std_logic;  -- write enable for memory
        -- report back once done
        z_limit_o:  out natural;    -- number of remaining elements (minus one)
        sum_done_o: out std_logic
    );
end entity circulant_sum;

architecture rtl of circulant_sum is

    constant    INVALID_POS:        pos:=       to_unsigned(P, position_bits);

    signal  i:              natural;
    signal  j:              natural;
    signal  k:              natural;
    signal  next_i:         natural;
    signal  next_j:         natural;
    signal  next_k:         natural;

    signal  i_not_done:     std_logic;
    signal  j_not_done:     std_logic;
    signal  neither_done:   std_logic;

    type state_t is (IDLE, COMPARE, SKIP, COPY_I, COPY_J, FILL_INVALID, DONE);
    signal  state:          state_t;
    signal  next_state:     state_t;

begin

    i_not_done      <=  '1' when i < a_limit_i and a_i /= INVALID_POS else
                        '0';
    j_not_done      <=  '1' when j < b_limit_i and b_i /= INVALID_POS else
                        '0';
    neither_done    <=  i_not_done and j_not_done;

    a_addr_o    <=  a_base_i + i;
    b_addr_o    <=  b_base_i + j;
    z_addr_o    <=  z_base_i + k;

    z_limit_o   <=  k;

    state_comb: process(
        state, start_i, i_not_done, j_not_done, neither_done,
        k, a_i, b_i
    )
    begin
        case state is
            when IDLE =>
                if start_i = '1' then
                    next_state  <=  COMPARE;
                else
                    next_state  <=  IDLE;
                end if;
            when COMPARE =>
                if neither_done = '1' and a_i = b_i then
                    next_state  <=  SKIP;
                elsif neither_done = '1' and a_i < b_i then
                    next_state  <=  COPY_I;
                elsif neither_done = '1' and a_i > b_i then
                    next_state  <=  COPY_J;
                elsif i_not_done = '1' and not j_not_done = '1' then
                    next_state  <=  COPY_I;
                elsif j_not_done = '1' and i_not_done = '0' then
                    next_state  <=  COPY_J;
                elsif i_not_done = '0' and j_not_done = '0'  -- cont.
                        and k < sum(M)*DV then
                    next_state  <=  FILL_INVALID;
                else
                    next_state  <=  DONE;
                end if;
            when SKIP =>
                next_state  <=  COMPARE;
            when COPY_I =>
                next_state  <=  COMPARE;
            when COPY_J =>
                next_state  <=  COMPARE;
            when FILL_INVALID =>
                next_state  <=  COMPARE;
            when DONE =>
                if start_i = '0' then
                    next_state  <=  IDLE;
                else
                    next_state  <=  DONE;
                end if;
            when others =>
                next_state  <=  IDLE;
        end case;
    end process state_comb;

    state_seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            state   <=  IDLE;
        elsif rising_edge(clk_i) then
            state   <=   next_state;
        end if;
    end process state_seq;

    output_comb: process(state, i, j, k, a_i, b_i)
    begin
        case state is
            when IDLE =>
                -- internal signals
                next_i      <=  0;
                next_j      <=  0;
                next_k      <=  0;
                -- outputs
                z_o         <=  INVALID_POS;
                wr_o        <=  '0';
                sum_done_o  <=  '0';
            when COMPARE =>
                -- internal signals
                next_i      <=  i;
                next_j      <=  j;
                next_k      <=  k;
                -- outputs
                z_o         <=  INVALID_POS;
                wr_o        <=  '0';
                sum_done_o  <=  '0';
            when SKIP =>
                -- internal signals
                next_i      <=  i + 1;
                next_j      <=  j + 1;
                next_k      <=  k;
                -- outputs
                z_o         <=  INVALID_POS;
                wr_o        <=  '0';
                sum_done_o  <=  '0';
            when COPY_I =>
                -- internal signals
                next_i      <=  i + 1;
                next_j      <=  j;
                next_k      <=  k + 1;
                -- outputs
                z_o         <=  a_i;
                wr_o        <=  '1';
                sum_done_o  <=  '0';
            when COPY_J =>
                -- internal signals
                next_i      <=  i;
                next_j      <=  j + 1;
                next_k      <=  k + 1;
                -- outputs
                z_o         <=  b_i;
                wr_o        <=  '1';
                sum_done_o  <=  '0';
            when FILL_INVALID =>
                -- internal signals
                next_i      <=  i;
                next_j      <=  j;
                next_k      <=  k + 1;
                -- outputs
                z_o         <=  INVALID_POS;
                wr_o        <=  '1';
                sum_done_o  <=  '0';
            when DONE =>
                -- internal signals
                next_i      <=  i;
                next_j      <=  j;
                next_k      <=  k + 1;
                -- outputs
                z_o         <=  INVALID_POS;
                wr_o        <=  '0';
                sum_done_o  <=  '1';
            when others =>
                -- internal signals
                next_i      <=  0;
                next_j      <=  0;
                next_k      <=  0;
                -- outputs
                z_o         <=  INVALID_POS;
                wr_o        <=  '0';
                sum_done_o  <=  '0';
        end case;
    end process output_comb;

    output_seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            i           <=  0;
            j           <=  0;
            k           <=  0;
        elsif rising_edge(clk_i) then
            i           <=  next_i;
            j           <=  next_j;
            k           <=  next_k;
        end if;
    end process output_seq;

end architecture rtl;
