library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.system_params.all;
    use work.matrix_types.all;
    use work.matrix_mult_functions.all;
    use work.memory_map.all;

entity key_reconstruction is
    port (
        -- control signals
        clk_i:          in  std_logic;
        rst_n_i:        in  std_logic;
        start_i:        in  std_logic;
        -- data, addresses and controls to memory
        a_i:            in  pos;
        a_addr_o:       out addr;
        b_i:            in  pos;
        b_addr_o:       out addr;
        z_o:            out pos;
        z_addr_o:       out addr;
        wr_o:           out std_logic;
        -- report back once done
        key_rec_done_o: out std_logic
    );
end entity key_reconstruction;

architecture structural of key_reconstruction is

    component key_reconstruction_sm is
        port(
            clk_i:              in  std_logic;  -- clock
            rst_n_i:            in  std_logic;  -- reset, active low
            start_i:            in  std_logic;  -- start key reconstruction when high
            -- circulant multiplication
            mult_start_o:       out std_logic;  -- start circulant multiplication
            mult_a_limit_o:     out natural;    -- element count (minus one) for matrix A
            mult_a_base_o:      out addr;       -- memory address of matrix A
            mult_b_limit_o:     out natural;    -- element count (minus one) for matrix B
            mult_b_base_o:      out addr;       -- memory address of matrix B
            mult_z_base_o:      out addr;       -- memory address of matrix Z
            mult_a_o:           out pos;        -- operand from 1st matrix
            mult_a_addr_i:      in  addr;       -- address for 1st matrix
            mult_b_o:           out pos;        -- operand from 2nd matrix
            mult_b_addr_i:      in  addr;       -- address for 2nd matrix
            mult_z_i:           in  pos;        -- Ltr result
            mult_z_addr_i:      in  addr;       -- address for Ltr result
            mult_wr_i:          in  std_logic;  -- write enable for memory
            mult_z_limit_i:     in  natural;    -- element count (minus one) of matrix Z
            mult_done_i:        in  std_logic;  -- circulant multiplication done
            -- sort
            sort_start_o:       out std_logic;  -- start sorting result of previous step
            sort_limit_o:       out natural;    -- element count (minus one) of stuff to sort
            sort_base_o:        out addr;       -- memory address of stuff to sort
            sort_a_o:           out pos;        -- element A from array
            sort_a_addr_i:      in  addr;       -- address of element A
            sort_b_o:           out pos;        -- element B from array
            sort_b_addr_i:      in  addr;       -- address of element B
            sort_z_i:           in  pos;        -- element Z to array
            sort_z_addr_i:      in  addr;       -- address of element Z
            sort_wr_i:          in  std_logic;  -- write enable for memory
            sort_done_i:        in  std_logic;  -- sorting done
            -- simplify
            simp_start_o:       out std_logic;  -- start simplifying sorted result
            simp_limit_o:       out natural;    -- element count (minus one) of stuff to simplify
            simp_base_o:        out addr;       -- memory address of stuff to simplify
            simp_a_o:           out pos;        -- first operand
            simp_a_addr_i:      in  addr;       -- address of first operand
            simp_b_o:           out pos;        -- second operand
            simp_b_addr_i:      in  addr;       -- address of second operand
            simp_z_i:           in  pos;        -- value (smaller then greater over two cycles)
            simp_z_addr_i:      in  addr;       -- position of ordered value
            simp_wr_i:          in  std_logic;  -- write enable for memory
            simp_z_limit_i:     in  natural;    -- element count (minus one) of simplified result
            simp_done_i:        in  std_logic;  -- simplifying done
            -- sum
            sum_start_o:        out std_logic;  -- start summing new result with partial result
            sum_a_limit_o:      out natural;
            sum_a_base_o:       out addr;
            sum_b_limit_o:      out natural;
            sum_b_base_o:       out addr;
            sum_z_base_o:       out addr;
            sum_a_o:            out pos;
            sum_a_addr_i:       in  addr;
            sum_b_o:            out pos;
            sum_b_addr_i:       in  addr;
            sum_z_i:            in  pos;
            sum_z_addr_i:       in  addr;
            sum_wr_i:           in  std_logic;  -- write enable for memory
            sum_z_limit_i:      in  natural;
            sum_done_i:         in  std_logic;
            -- copy
            copy_start_o:       out std_logic;
            copy_limit_o:       out natural;
            copy_a_base_o:      out addr;
            copy_z_base_o:      out addr;
            copy_a_o:           out pos;        -- element A from source array
            copy_a_addr_i:      in  addr;       -- address of element A
            copy_z_i:           in  pos;        -- element Z to destination array
            copy_z_addr_i:      in  addr;       -- address of element Z
            copy_wr_i:          in  std_logic;  -- write enable for memory
            copy_done_i:        in  std_logic;
            -- controls to memory
            a_i:                in  pos;
            a_addr_o:           out addr;
            b_i:                in  pos;
            b_addr_o:           out addr;
            z_o:                out pos;
            z_addr_o:           out addr;
            wr_o:               out std_logic;
            -- report back once done
            key_rec_done_o:     out std_logic
        );
    end component key_reconstruction_sm;

    component circulant_multiplication is
        port(
            -- control signals
            clk_i:          in  std_logic;
            rst_n_i:        in  std_logic;
            start_i:        in  std_logic;
            -- function arguments in (not latched)
            a_limit_i:      in  natural;    -- number of elements in 1st matrix
            a_base_i:       in  addr;       -- base address of 1st matrix
            b_limit_i:      in  natural;    -- number of elements in 2nd matrix
            b_base_i:       in  addr;       -- base address of 2nd matrix
            z_base_i:       in  addr;       -- base address of result matrix
            -- data, addresses and controls to memory
            a_i:            in  pos;        -- operand from 1st matrix
            a_addr_o:       out addr;       -- address for 1st matrix
            b_i:            in  pos;        -- operand from 2nd matrix
            b_addr_o:       out addr;       -- address for 2nd matrix
            z_o:            out pos;        -- Ltr result
            z_addr_o:       out addr;       -- address for Ltr result
            wr_o:           out std_logic;  -- write enable for memory
            -- report back once done
            z_limit_o:      out natural;    -- number of elements in result matrix (minus one)
            mult_done_o:    out std_logic   -- high when done
        );
    end component circulant_multiplication;

    component sort is
        port(
            -- control logic
            clk_i:          in  std_logic;
            rst_n_i:        in  std_logic;
            start_i:        in  std_logic;
            -- function arguments in (not latched)
            limit_i:        in  natural;    -- number of elements in array to sort
            base_i:         in  addr;       -- base addr of array to sort
            -- data, addresses and controls to memory
            a_i:            in  pos;        -- element A from array
            a_addr_o:       out addr;       -- address of element A
            b_i:            in  pos;        -- element B from array
            b_addr_o:       out addr;       -- address of element B
            z_o:            out pos;        -- element Z to array
            z_addr_o:       out addr;       -- address of element Z
            wr_o:           out std_logic;  -- write enable for memory
            -- report back once done
            sort_done_o:    out std_logic   -- high when done
        );
    end component sort;

    component simplify is
        port(
            -- control logic
            clk_i:          in  std_logic;
            rst_n_i:        in  std_logic;
            start_i:        in  std_logic;
            -- function arguments in (not latched)
            limit_i:        in  natural;    -- number of elements in array to simplify
            base_i:         in  addr;       -- base addr of array to sort
            -- data, addresses and controls to memory
            a_i:            in  pos;        -- first operand
            a_addr_o:       out addr;       -- address of first operand
            b_i:            in  pos;        -- second operand
            b_addr_o:       out addr;       -- address of second operand
            z_o:            out pos;        -- value (smaller then greater over two cycles)
            z_addr_o:       out addr;       -- position of ordered value
            wr_o:           out std_logic;  -- write enable for memory
            -- report back once done
            z_limit_o:      out natural;    -- number of remaining elements (minus one)
            simp_done_o:    out std_logic
        );
    end component simplify;

    component circulant_sum is
        port(
            -- control signals
            clk_i:      in  std_logic;
            rst_n_i:    in  std_logic;
            start_i:    in  std_logic;
            -- function arguments in (not latched)
            a_limit_i:  in  natural;    -- number of elements in 1st matrix
            a_base_i:   in  addr;       -- base address of 1st matrix
            b_limit_i:  in  natural;    -- number of elements in 2nd matrix
            b_base_i:   in  addr;       -- base address of 2nd matrix
            z_base_i:   in  addr;       -- base address of result matrix
            -- data, addresses and controls to memory
            a_i:        in  pos;
            a_addr_o:   out addr;
            b_i:        in  pos;
            b_addr_o:   out addr;
            z_o:        out pos;
            z_addr_o:   out addr;
            wr_o:       out std_logic;  -- write enable for memory
            -- report back once done
            z_limit_o:  out natural;    -- number of remaining elements (minus one)
            sum_done_o: out std_logic
        );
    end component circulant_sum;

    component mem_copy is
        port (
            -- control signals
            clk_i:          in  std_logic;
            rst_n_i:        in  std_logic;
            start_i:        in  std_logic;
            -- function arguments in (not latched)
            limit_i:        in  natural;    -- number of elements (minus one) in source array
            a_base_i:       in  addr;       -- base addr of source array
            z_base_i:       in  addr;       -- base addr of destination array
            -- data, addresses and controls to memory
            a_i:            in  pos;        -- element A from source array
            a_addr_o:       out addr;       -- address of element A
            z_o:            out pos;        -- element Z to destination array
            z_addr_o:       out addr;       -- address of element Z
            wr_o:           out std_logic;  -- write enable for memory
            -- report back once done
            copy_done_o:    out std_logic   -- high when done
        );
    end component mem_copy;

    signal  mult_start_w:   std_logic;
    signal  mult_a_limit_w: natural;
    signal  mult_a_base_w:  addr;
    signal  mult_b_limit_w: natural;
    signal  mult_b_base_w:  addr;
    signal  mult_z_base_w:  addr;
    signal  mult_a_w:       pos;
    signal  mult_a_addr_w:  addr;
    signal  mult_b_w:       pos;
    signal  mult_b_addr_w:  addr;
    signal  mult_z_w:       pos;
    signal  mult_z_addr_w:  addr;
    signal  mult_wr_w:      std_logic;
    signal  mult_z_limit_w: natural;
    signal  mult_done_w:    std_logic;
    signal  sort_start_w:   std_logic;
    signal  sort_limit_w:   natural;
    signal  sort_base_w:    addr;
    signal  sort_a_w:       pos;
    signal  sort_a_addr_w:  addr;
    signal  sort_b_w:       pos;
    signal  sort_b_addr_w:  addr;
    signal  sort_z_w:       pos;
    signal  sort_z_addr_w:  addr;
    signal  sort_wr_w:      std_logic;
    signal  sort_done_w:    std_logic;
    signal  simp_start_w:   std_logic;
    signal  simp_limit_w:   natural;
    signal  simp_base_w:    addr;
    signal  simp_a_w:       pos;
    signal  simp_a_addr_w:  addr;
    signal  simp_b_w:       pos;
    signal  simp_b_addr_w:  addr;
    signal  simp_z_w:       pos;
    signal  simp_z_addr_w:  addr;
    signal  simp_wr_w:      std_logic;
    signal  simp_z_limit_w: natural;
    signal  simp_done_w:    std_logic;
    signal  sum_start_w:    std_logic;
    signal  sum_a_limit_w:  natural;
    signal  sum_a_base_w:   addr;
    signal  sum_b_limit_w:  natural;
    signal  sum_b_base_w:   addr;
    signal  sum_z_base_w:   addr;
    signal  sum_a_w:        pos;
    signal  sum_a_addr_w:   addr;
    signal  sum_b_w:        pos;
    signal  sum_b_addr_w:   addr;
    signal  sum_z_w:        pos;
    signal  sum_z_addr_w:   addr;
    signal  sum_wr_w:       std_logic;
    signal  sum_z_limit_w:  natural;
    signal  sum_done_w:     std_logic;
    signal  copy_start_w:   std_logic;
    signal  copy_limit_w:   natural;
    signal  copy_a_base_w:  addr;
    signal  copy_z_base_w:  addr;
    signal  copy_a_w:       pos;
    signal  copy_a_addr_w:  addr;
    signal  copy_z_w:       pos;
    signal  copy_z_addr_w:  addr;
    signal  copy_wr_w:      std_logic;
    signal  copy_done_w:    std_logic;

begin

    KRSM: key_reconstruction_sm
        port map(
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  start_i,
            -- circulant multiplication
            mult_start_o    =>  mult_start_w,
            mult_a_limit_o  =>  mult_a_limit_w,
            mult_a_base_o   =>  mult_a_base_w,
            mult_b_limit_o  =>  mult_b_limit_w,
            mult_b_base_o   =>  mult_b_base_w,
            mult_z_base_o   =>  mult_z_base_w,
            mult_a_o        =>  mult_a_w,
            mult_a_addr_i   =>  mult_a_addr_w,
            mult_b_o        =>  mult_b_w,
            mult_b_addr_i   =>  mult_b_addr_w,
            mult_z_i        =>  mult_z_w,
            mult_z_addr_i   =>  mult_z_addr_w,
            mult_wr_i       =>  mult_wr_w,
            mult_z_limit_i  =>  mult_z_limit_w,
            mult_done_i     =>  mult_done_w,
            -- sort
            sort_start_o    =>  sort_start_w,
            sort_limit_o    =>  sort_limit_w,
            sort_base_o     =>  sort_base_w,
            sort_a_o        =>  sort_a_w,
            sort_a_addr_i   =>  sort_a_addr_w,
            sort_b_o        =>  sort_b_w,
            sort_b_addr_i   =>  sort_b_addr_w,
            sort_z_i        =>  sort_z_w,
            sort_z_addr_i   =>  sort_z_addr_w,
            sort_wr_i       =>  sort_wr_w,
            sort_done_i     =>  sort_done_w,
            -- simplify
            simp_start_o    =>  simp_start_w,
            simp_limit_o    =>  simp_limit_w,
            simp_base_o     =>  simp_base_w,
            simp_a_o        =>  simp_a_w,
            simp_a_addr_i   =>  simp_a_addr_w,
            simp_b_o        =>  simp_b_w,
            simp_b_addr_i   =>  simp_b_addr_w,
            simp_z_i        =>  simp_z_w,
            simp_z_addr_i   =>  simp_z_addr_w,
            simp_wr_i       =>  simp_wr_w,
            simp_z_limit_i  =>  simp_z_limit_w,
            simp_done_i     =>  simp_done_w,
            -- sum
            sum_start_o     =>  sum_start_w,
            sum_a_limit_o   =>  sum_a_limit_w,
            sum_a_base_o    =>  sum_a_base_w,
            sum_b_limit_o   =>  sum_b_limit_w,
            sum_b_base_o    =>  sum_b_base_w,
            sum_z_base_o    =>  sum_z_base_w,
            sum_a_o         =>  sum_a_w,
            sum_a_addr_i    =>  sum_a_addr_w,
            sum_b_o         =>  sum_b_w,
            sum_b_addr_i    =>  sum_b_addr_w,
            sum_z_i         =>  sum_z_w,
            sum_z_addr_i    =>  sum_z_addr_w,
            sum_wr_i        =>  sum_wr_w,
            sum_z_limit_i   =>  sum_z_limit_w,
            sum_done_i      =>  sum_done_w,
            -- copy
            copy_start_o    =>  copy_start_w,
            copy_limit_o    =>  copy_limit_w,
            copy_a_base_o   =>  copy_a_base_w,
            copy_z_base_o   =>  copy_z_base_w,
            copy_a_o        =>  copy_a_w,
            copy_a_addr_i   =>  copy_a_addr_w,
            copy_z_i        =>  copy_z_w,
            copy_z_addr_i   =>  copy_z_addr_w,
            copy_wr_i       =>  copy_wr_w,
            copy_done_i     =>  copy_done_w,
            -- controls to memory
            a_i             =>  a_i,
            a_addr_o        =>  a_addr_o,
            b_i             =>  b_i,
            b_addr_o        =>  b_addr_o,
            z_o             =>  z_o,
            z_addr_o        =>  z_addr_o,
            wr_o            =>  wr_o,
            -- report back once done
            key_rec_done_o  =>  key_rec_done_o
        );

    CM: circulant_multiplication
        port map(
            -- control signals
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  mult_start_w,
            -- function arguments in (not latched)
            a_limit_i       =>  mult_a_limit_w,
            a_base_i        =>  mult_a_base_w,
            b_limit_i       =>  mult_b_limit_w,
            b_base_i        =>  mult_b_base_w,
            z_base_i        =>  mult_z_base_w,
            -- data, addresses and controls to memory
            a_i             =>  mult_a_w,
            a_addr_o        =>  mult_a_addr_w,
            b_i             =>  mult_b_w,
            b_addr_o        =>  mult_b_addr_w,
            z_o             =>  mult_z_w,
            z_addr_o        =>  mult_z_addr_w,
            wr_o            =>  mult_wr_w,
            -- report back once done
            z_limit_o       =>  mult_z_limit_w,
            mult_done_o     =>  mult_done_w
        );

    SO: sort
        port map(
            -- control logic
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  sort_start_w,
            -- function arguments in (not latched)
            limit_i         =>  sort_limit_w,
            base_i          =>  sort_base_w,
            -- data, addresses and controls to memory
            a_i             =>  sort_a_w,
            a_addr_o        =>  sort_a_addr_w,
            b_i             =>  sort_b_w,
            b_addr_o        =>  sort_b_addr_w,
            z_o             =>  sort_z_w,
            z_addr_o        =>  sort_z_addr_w,
            wr_o            =>  sort_wr_w,
            -- report back once done
            sort_done_o     =>  sort_done_w
        );

    SIMP: simplify
        port map(
            -- control logic
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  simp_start_w,
            -- function arguments in (not latched)
            limit_i         =>  simp_limit_w,
            base_i          =>  simp_base_w,
            -- data, addresses and controls to memory
            a_i             =>  simp_a_w,
            a_addr_o        =>  simp_a_addr_w,
            b_i             =>  simp_b_w,
            b_addr_o        =>  simp_b_addr_w,
            z_o             =>  simp_z_w,
            z_addr_o        =>  simp_z_addr_w,
            wr_o            =>  simp_wr_w,
            -- report back once done
            z_limit_o       =>  simp_z_limit_w,
            simp_done_o     =>  simp_done_w
        );

    CS: circulant_sum
        port map(
            -- control signals
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  sum_start_w,
            -- function arguments in (not latched)
            a_limit_i       =>  sum_a_limit_w,
            a_base_i        =>  sum_a_base_w,
            b_limit_i       =>  sum_b_limit_w,
            b_base_i        =>  sum_b_base_w,
            z_base_i        =>  sum_z_base_w,
            -- data, addresses and controls to memory
            a_i             =>  sum_a_w,
            a_addr_o        =>  sum_a_addr_w,
            b_i             =>  sum_b_w,
            b_addr_o        =>  sum_b_addr_w,
            z_o             =>  sum_z_w,
            z_addr_o        =>  sum_z_addr_w,
            wr_o            =>  sum_wr_w,
            -- report back once done
            z_limit_o       =>  sum_z_limit_w,
            sum_done_o      =>  sum_done_w
        );

    CP: mem_copy
        port map(
            -- control signals
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  copy_start_w,
            -- function arguments in (not latched)
            limit_i         =>  copy_limit_w,
            a_base_i        =>  copy_a_base_w,
            z_base_i        =>  copy_z_base_w,
            -- data, addresses and controls to memory
            a_i             =>  copy_a_w,
            a_addr_o        =>  copy_a_addr_w,
            z_o             =>  copy_z_w,
            z_addr_o        =>  copy_z_addr_w,
            wr_o            =>  copy_wr_w,
            -- report back once done
            copy_done_o     =>  copy_done_w
        );

end architecture;
