library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.system_params.all;
    use work.matrix_types.all;
    use work.memory_map.all;
    use work.matrix_mult_functions.all;

entity key_reconstruction_sm is
    port(
        clk_i:              in  std_logic;  -- clock
        rst_n_i:            in  std_logic;  -- reset, active low
        start_i:            in  std_logic;  -- start key reconstruction when high
        -- circulant multiplication
        mult_start_o:       out std_logic;  -- start circulant multiplication
        mult_a_limit_o:     out natural;    -- element count (minus one) for matrix A
        mult_a_base_o:      out addr;       -- memory address of matrix A
        mult_b_limit_o:     out natural;    -- element count (minus one) for matrix B
        mult_b_base_o:      out addr;       -- memory address of matrix B
        mult_z_base_o:      out addr;       -- memory address of matrix Z
        mult_a_o:           out pos;        -- operand from 1st matrix
        mult_a_addr_i:      in  addr;       -- address for 1st matrix
        mult_b_o:           out pos;        -- operand from 2nd matrix
        mult_b_addr_i:      in  addr;       -- address for 2nd matrix
        mult_z_i:           in  pos;        -- Ltr result
        mult_z_addr_i:      in  addr;       -- address for Ltr result
        mult_wr_i:          in  std_logic;  -- write enable for memory
        mult_z_limit_i:     in  natural;    -- element count (minus one) of matrix Z
        mult_done_i:        in  std_logic;  -- circulant multiplication done
        -- sort
        sort_start_o:       out std_logic;  -- start sorting result of previous step
        sort_limit_o:       out natural;    -- element count (minus one) of stuff to sort
        sort_base_o:        out addr;       -- memory address of stuff to sort
        sort_a_o:           out pos;        -- element A from array
        sort_a_addr_i:      in  addr;       -- address of element A
        sort_b_o:           out pos;        -- element B from array
        sort_b_addr_i:      in  addr;       -- address of element B
        sort_z_i:           in  pos;        -- element Z to array
        sort_z_addr_i:      in  addr;       -- address of element Z
        sort_wr_i:          in  std_logic;  -- write enable for memory
        sort_done_i:        in  std_logic;  -- sorting done
        -- simplify
        simp_start_o:       out std_logic;  -- start simplifying sorted result
        simp_limit_o:       out natural;    -- element count (minus one) of stuff to simplify
        simp_base_o:        out addr;       -- memory address of stuff to simplify
        simp_a_o:           out pos;        -- first operand
        simp_a_addr_i:      in  addr;       -- address of first operand
        simp_b_o:           out pos;        -- second operand
        simp_b_addr_i:      in  addr;       -- address of second operand
        simp_z_i:           in  pos;        -- value (smaller then greater over two cycles)
        simp_z_addr_i:      in  addr;       -- position of ordered value
        simp_wr_i:          in  std_logic;  -- write enable for memory
        simp_z_limit_i:     in  natural;    -- element count (minus one) of simplified result
        simp_done_i:        in  std_logic;  -- simplifying done
        -- sum
        sum_start_o:        out std_logic;  -- start summing new result with partial result
        sum_a_limit_o:      out natural;
        sum_a_base_o:       out addr;
        sum_b_limit_o:      out natural;
        sum_b_base_o:       out addr;
        sum_z_base_o:       out addr;
        sum_a_o:            out pos;
        sum_a_addr_i:       in  addr;
        sum_b_o:            out pos;
        sum_b_addr_i:       in  addr;
        sum_z_i:            in  pos;
        sum_z_addr_i:       in  addr;
        sum_wr_i:           in  std_logic;  -- write enable for memory
        sum_z_limit_i:      in  natural;
        sum_done_i:         in  std_logic;
        -- copy
        copy_start_o:       out std_logic;
        copy_limit_o:       out natural;
        copy_a_base_o:      out addr;
        copy_z_base_o:      out addr;
        copy_a_o:           out pos;        -- element A from source array
        copy_a_addr_i:      in  addr;       -- address of element A
        copy_z_i:           in  pos;        -- element Z to destination array
        copy_z_addr_i:      in  addr;       -- address of element Z
        copy_wr_i:          in  std_logic;  -- write enable for memory
        copy_done_i:        in  std_logic;
        -- controls to memory
        a_i:                in  pos;
        a_addr_o:           out addr;
        b_i:                in  pos;
        b_addr_o:           out addr;
        z_o:                out pos;
        z_addr_o:           out addr;
        wr_o:               out std_logic;
        -- report back once done
        key_rec_done_o:     out std_logic
    );
end entity key_reconstruction_sm;

architecture rtl of key_reconstruction_sm is

    signal  mul_result_size_r:  natural;
    signal  mul_result_size_c:  natural;
    signal  simp_result_size_r: natural;
    signal  simp_result_size_c: natural;
    signal  sum_result_size_r:  natural;
    signal  sum_result_size_c:  natural;

    type state_t is (IDLE, CIRC_MUL, SORT, SIMPLIFY, CIRC_SUM, COPY, DONE);
    signal  state:      state_t;
    signal  next_state: state_t;

    -- keep track of which blocks we are processing (rows and cols of Qtr)
    signal  block_row_c:        natural;
    signal  block_col_c:        natural;
    signal  block_row_r:        natural;
    signal  block_col_r:        natural;

    -- base addresses of circulant blocks of Htr and Qtr, plus element count for
    -- circulant block of Qtr (not constant)
    signal  Htr_block_base:     addr;
    signal  Qtr_block_base:     addr;
    signal  Htr_block_count:    natural;
    signal  Qtr_block_count:    natural;

    constant    Qtr_block_limit_map:    n_matrix:=      get_block_limits;
    constant    Htr_block_base_map:     addr_array:=    get_Htr_block_base;
    constant    Qtr_block_base_map:     addr_matrix:=   get_Qtr_block_base;

begin

    Htr_block_count <=  DV - 1; -- compile-time parameter
    Qtr_block_count <=  Qtr_block_limit_map(block_row_r, block_col_r);

    -- Htr only has block rows that are multiplyied with the corresponding
    -- block column of Qtr. Since computing it makes little sense, we just go
    -- with a small LUT.
    Htr_block_base  <=  Htr_block_base_map(block_col_r);
    Qtr_block_base  <=  Qtr_block_base_map(block_row_r, block_col_r);

    -- Circulant multiplication: Qtr*Htr = Ltr (as L = H*Q)
    -- A matrix is thus a block from Qtr, B matrix is a block from Htr
    mult_a_limit_o  <=  Htr_block_count;
    mult_a_base_o   <=  HTR_BASE_ADDR + Htr_block_base;
    mult_b_limit_o  <=  Qtr_block_count;
    mult_b_base_o   <=  QTR_BASE_ADDR + Qtr_block_base;
    mult_z_base_o   <=  MUL_RES_BASE_ADDR;
    mult_a_o        <=  a_i;
    mult_b_o        <=  b_i;

    -- Sorting (done in place): having the exact count could save huge effort,
    -- thus it makes sense to compute it. It makes even more sense to have the
    -- count come out from the circulant multiplication block so we do not need
    -- any multiplier.
    -- Done, count now comes from circulant multiplication block.
    sort_limit_o    <=  mul_result_size_r;
    sort_base_o     <=  MUL_RES_BASE_ADDR;
    sort_a_o        <=  a_i;
    sort_b_o        <=  b_i;

    -- Simplification (done in place): as above.
    simp_limit_o    <=  mul_result_size_r;
    simp_base_o     <=  MUL_RES_BASE_ADDR;
    simp_a_o        <=  a_i;
    simp_b_o        <=  b_i;

    -- Circulant sum (needs a tmp address to avoid overwriting an operand):
    -- from here on exact size depends on runtime operands. Simplify block knows
    -- the exact number (likely the maximum one to be honest, but we already
    -- have a counter to generate addresses so we can use that with some
    -- register overhead).
    -- Done, count now comes from simplify block.
    -- Keep track of growing size of Ltr? Using max number seems overkill for
    -- the first few iterations. TODO.
    sum_a_limit_o   <=  simp_result_size_r;
    sum_a_base_o    <=  MUL_RES_BASE_ADDR;
    sum_b_limit_o   <=  sum_result_size_r;
    sum_b_base_o    <=  LTR_BASE_ADDR + block_row_r*sum(M)*DV;
    sum_z_base_o    <=  SUM_TMP_BASE_ADDR;
    sum_a_o         <=  a_i;
    sum_b_o         <=  b_i;

    -- Copy
    copy_limit_o    <=  sum_result_size_r;
    copy_a_base_o   <=  MUL_RES_BASE_ADDR when block_col_r = 0 else
                        SUM_TMP_BASE_ADDR;
    copy_z_base_o   <=  LTR_BASE_ADDR + block_row_r*sum(M)*DV;
    copy_a_o        <=  a_i;

    state_comb: process(
        state, start_i, mult_done_i, sort_done_i, simp_done_i, sum_done_i,
        copy_done_i, block_row_r, block_col_r
    )
    begin
        case state is
            when IDLE =>
                if start_i = '1' then
                    next_state  <=  CIRC_MUL;
                else
                    next_state  <=  IDLE;
                end if;
            when CIRC_MUL =>
                if mult_done_i = '1' then
                    next_state  <=  SORT;
                else
                    next_state  <=  CIRC_MUL;
                end if;
            when SORT =>
                if sort_done_i = '1' then
                    next_state  <=  SIMPLIFY;
                else
                    next_state  <=  SORT;
                end if;
            when SIMPLIFY =>
                if simp_done_i = '1' then
                    if block_col_r = 0 then
                        -- first block of row adds to nothing, skip (but copy!)
                        next_state  <=  COPY;
                    else
                        next_state  <=  CIRC_SUM;
                    end if;
                else
                    next_state  <=  SIMPLIFY;
                end if;
            when CIRC_SUM =>
                if sum_done_i = '1' then
                    next_state  <=  COPY;
                else
                    next_state  <=  CIRC_SUM;
                end if;
            when COPY =>
                if copy_done_i = '1' then
                    if block_row_r = N0-1 and block_col_r = N0-1 then
                        next_state  <=  DONE;
                    else
                        next_state  <=  CIRC_MUL;
                    end if;
                else
                    next_state  <=  COPY;
                end if;
            when DONE =>
                if start_i = '0' then
                    next_state  <=  IDLE;
                else
                    next_state  <=  DONE;
                end if;
            when others =>
                next_state  <=  IDLE;
        end case;
    end process state_comb;

    state_seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            state   <=  IDLE;
        elsif rising_edge(clk_i) then
            state   <=  next_state;
        end if;
    end process state_seq;

    output_comb: process(
        state, block_row_r, block_col_r, mul_result_size_r, mult_z_limit_i,
        simp_z_limit_i, sum_done_i,
        mult_z_i, mult_a_addr_i, mult_b_addr_i, mult_z_addr_i, mult_wr_i,
        sort_z_i, sort_a_addr_i, sort_b_addr_i, sort_z_addr_i, sort_wr_i,
        simp_z_i, simp_a_addr_i, simp_b_addr_i, simp_z_addr_i, simp_wr_i,
        sum_z_i, sum_a_addr_i, sum_b_addr_i, sum_z_addr_i, sum_wr_i,
        copy_z_i, copy_a_addr_i, copy_z_addr_i, copy_wr_i
    )
    begin
        block_row_c         <=  block_row_r;
        block_col_c         <=  block_col_r;
        mul_result_size_c   <=  mul_result_size_r;
        simp_result_size_c  <=  simp_result_size_r;
        if block_col_r = 0 then
            -- no sum for first block of row!
            sum_result_size_c   <=  simp_result_size_r;
        else
            sum_result_size_c   <=  sum_result_size_r;
        end if;
        mult_start_o        <=  '0';
        sort_start_o        <=  '0';
        simp_start_o        <=  '0';
        sum_start_o         <=  '0';
        copy_start_o        <=  '0';
        key_rec_done_o      <=  '0';
        z_o                 <=  mult_z_i;
        a_addr_o            <=  mult_a_addr_i;
        b_addr_o            <=  mult_b_addr_i;
        z_addr_o            <=  mult_z_addr_i;
        wr_o                <=  mult_wr_i;
        case state  is
            when IDLE =>
                block_row_c         <=  0;
                block_col_c         <=  0;
            when CIRC_MUL =>
                z_o                 <=  mult_z_i;
                a_addr_o            <=  mult_a_addr_i;
                b_addr_o            <=  mult_b_addr_i;
                z_addr_o            <=  mult_z_addr_i;
                wr_o                <=  mult_wr_i;
                mult_start_o        <=  '1';
                -- This flips a lot, but we don't need extra gates to avoid
                -- its flipping. Change this if power is a concern (likely
                -- if this ever makes it to silicon).
                mul_result_size_c   <=  mult_z_limit_i;
            when SORT =>
                z_o                 <=  sort_z_i;
                a_addr_o            <=  sort_a_addr_i;
                b_addr_o            <=  sort_b_addr_i;
                z_addr_o            <=  sort_z_addr_i;
                wr_o                <=  sort_wr_i;
                sort_start_o        <=  '1';
            when SIMPLIFY =>
                z_o                 <=  simp_z_i;
                a_addr_o            <=  simp_a_addr_i;
                b_addr_o            <=  simp_b_addr_i;
                z_addr_o            <=  simp_z_addr_i;
                wr_o                <=  simp_wr_i;
                simp_start_o        <=  '1';
                -- This flips a lot, but we don't need extra gates to avoid
                -- its flipping. Change this if power is a concern (likely
                -- if this ever makes it to silicon).
                simp_result_size_c   <=  simp_z_limit_i;
            when CIRC_SUM =>
                z_o                 <=  sum_z_i;
                a_addr_o            <=  sum_a_addr_i;
                b_addr_o            <=  sum_b_addr_i;
                z_addr_o            <=  sum_z_addr_i;
                wr_o                <=  sum_wr_i;
                sum_start_o         <=  '1';
                if sum_done_i = '1' then
                    sum_result_size_c   <=  sum_z_limit_i;
                end if;
            when COPY =>
                z_o                 <=  copy_z_i;
                a_addr_o            <=  copy_a_addr_i;
                z_addr_o            <=  copy_z_addr_i;
                wr_o                <=  copy_wr_i;
                copy_start_o        <=  '1';
                if copy_done_i = '1' then
                    if block_col_r = N0-1 then
                        if block_row_r /= N0-1 then
                            block_row_c <=  block_row_r + 1;
                            block_col_c <=  0;
                        end if;
                    else
                        block_row_c <=  block_row_r;
                        block_col_c <=  block_col_r + 1;
                    end if;
                end if;
            when DONE =>
                key_rec_done_o  <=  '1';
            when others =>
                block_row_c <=  0;
                block_col_c <=  0;
        end case;
    end process output_comb;

    output_seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            block_row_r         <=  0;
            block_col_r         <=  0;
            mul_result_size_r   <=  0;
            simp_result_size_r  <=  0;
            sum_result_size_r   <=  0;
        elsif rising_edge(clk_i) then
            block_row_r         <=  block_row_c;
            block_col_r         <=  block_col_c;
            mul_result_size_r   <=  mul_result_size_c;
            simp_result_size_r  <=  simp_result_size_c;
            sum_result_size_r   <=  sum_result_size_c;
        end if;
    end process output_seq;

end architecture;
