library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.matrix_types.all;

entity mem_copy is
    port (
        -- control signals
        clk_i:          in  std_logic;
        rst_n_i:        in  std_logic;
        start_i:        in  std_logic;
        -- function arguments in (not latched)
        limit_i:        in  natural;    -- number of elements in source array
        a_base_i:       in  addr;       -- base addr of source array
        z_base_i:       in  addr;       -- base addr of destination array
        -- data, addresses and controls to memory
        a_i:            in  pos;        -- element A from source array
        a_addr_o:       out addr;       -- address of element A
        z_o:            out pos;        -- element Z to destination array
        z_addr_o:       out addr;       -- address of element Z
        wr_o:           out std_logic;  -- write enable for memory
        -- report back once done
        copy_done_o:    out std_logic   -- high when done
    );
end entity mem_copy;

architecture rtl of mem_copy is

    signal  i:          natural;
    signal  next_i:     natural;

    type state_t is (IDLE, COPY, DONE);
    signal  state:      state_t;
    signal  next_state: state_t;

begin

    a_addr_o    <=  a_base_i + i;
    z_addr_o    <=  z_base_i + i;
    z_o         <=  a_i;

    state_comb: process(state, start_i, i, limit_i)
    begin
        case state is
            when IDLE =>
                if start_i = '1' then
                    next_state  <=  COPY;
                else
                    next_state  <=  IDLE;
                end if;
            when COPY =>
                if i = limit_i-1 then
                    next_state  <=  DONE;
                else
                    next_state  <=  COPY;
                end if;
            when DONE =>
                if start_i = '0' then
                    next_state  <=  IDLE;
                else
                    next_state  <=  DONE;
                end if;
            when others =>
                next_state  <=  IDLE;
        end case;
    end process state_comb;

    state_seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            state   <=  IDLE;
        elsif rising_edge(clk_i) then
            state   <=  next_state;
        end if;
    end process state_seq;

    output_comb: process(state, i)
    begin
        case state is
            when IDLE =>
                -- internal signals
                next_i      <=  0;
                -- outputs
                wr_o        <=  '0';
                copy_done_o <=  '0';
            when COPY =>
                -- internal signals
                next_i      <=  i + 1;
                -- outputs
                wr_o        <=  '1';
                copy_done_o <=  '0';
            when DONE =>
                -- internal signals
                next_i      <=  0;
                -- outputs
                wr_o        <=  '0';
                copy_done_o <=  '1';
            when others =>
                -- internal signals
                next_i      <=  i;
                -- outputs
                wr_o        <=  '0';
                copy_done_o <=  '0';
        end case;
    end process output_comb;

    output_seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            i   <=  0;
        elsif rising_edge(clk_i) then
            i   <=  next_i;
        end if;
    end process output_seq;

end architecture rtl;
