-- Future improvements: merge states, most of this can be done single cycle.
-- Combinatorics for the comparison is a xor followed by unary or, 0 if
-- different -> direct carryover to wr_o

library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.system_params.all;
    use work.matrix_types.all;
    use work.matrix_mult_functions.all;

entity simplify is
    port(
        -- control logic
        clk_i:          in  std_logic;
        rst_n_i:        in  std_logic;
        start_i:        in  std_logic;
        -- function arguments in (not latched)
        limit_i:        in  natural;    -- number of elements in array to simplify
        base_i:         in  addr;       -- base addr of array to sort
        -- data, addresses and controls to memory
        a_i:            in  pos;        -- first operand
        a_addr_o:       out addr;       -- address of first operand
        b_i:            in  pos;        -- second operand
        b_addr_o:       out addr;       -- address of second operand
        z_o:            out pos;        -- value (smaller then greater over two cycles)
        z_addr_o:       out addr;       -- position of ordered value
        wr_o:           out std_logic;  -- write enable for memory
        -- report back once done
        z_limit_o:      out natural;    -- number of remaining elements
        simp_done_o:    out std_logic
    );
end entity simplify;

architecture rtl of simplify is

    signal  i:          natural;
    signal  j:          natural;
    signal  next_i:     natural;
    signal  next_j:     natural;

    type state_t is (IDLE, CHECK_I, SKIP, COPY, UPDATE_I, LAST_COPY, CHECK_J, FILL_INVALID, DONE);
    signal  state:      state_t;
    signal  next_state: state_t;

    constant    INVALID_POS:    pos:=       to_unsigned(P, position_bits);

begin

    a_addr_o    <=  base_i + to_unsigned(i, address_bits);
    b_addr_o    <=  base_i + to_unsigned(i + 1, address_bits);
    z_addr_o    <=  base_i + to_unsigned(j, address_bits);
    z_limit_o   <=  j;

    state_comb: process(state, start_i, i, j, a_i, b_i, limit_i)
    begin
        case state is
            when IDLE =>
                if start_i = '1' then
                    next_state  <=  CHECK_I;
                else
                    next_state  <=  IDLE;
                end if;
            when CHECK_I =>
                if i < limit_i - 1 and a_i = b_i then
                    next_state  <=  SKIP;
                elsif i < limit_i - 1 and a_i /= b_i then
                    next_state  <=  COPY;
                elsif i = limit_i - 1 then
                    next_state  <=  LAST_COPY;
                else -- if i = limit_i
                    next_state  <=  CHECK_J;
                end if;
            when SKIP =>
                next_state  <=  UPDATE_I;
            when COPY =>
                next_state  <=  UPDATE_I;
            when UPDATE_I =>
                next_state  <=  CHECK_I;
            when LAST_COPY =>
                next_state  <=  CHECK_J;
            when CHECK_J    =>
                if j <= limit_i -1 then
                    next_state  <=  FILL_INVALID;
                else
                    next_state  <=  DONE;
                end if;
            when FILL_INVALID =>
                next_state  <=  CHECK_J;
            when DONE =>
                if start_i = '0' then
                    next_state  <=  IDLE;
                else
                    next_state  <=  DONE;
                end if;
            when others =>
                next_state  <=  IDLE;
        end case;
    end process state_comb;

    state_seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            state   <=  IDLE;
        elsif rising_edge(clk_i) then
            state   <=  next_state;
        end if;
    end process state_seq;

    output_comb: process(state, a_i, i, j)
    begin
        case state is
            when IDLE =>
                -- internal signals
                next_i          <=  0;
                next_j          <=  0;
                -- outputs
                z_o             <=  a_i;
                wr_o            <=  '0';
                simp_done_o     <=  '0';
            when CHECK_I =>
                -- internal signals
                next_i          <=  i;
                next_j          <=  j;
                -- outputs
                z_o             <=  a_i;
                wr_o            <=  '0';
                simp_done_o      <=  '0';
            when SKIP =>
                -- internal signals
                next_i          <=  i + 1;
                next_j          <=  j;
                -- outputs
                z_o             <=  a_i;
                wr_o            <=  '0';
                simp_done_o     <=  '0';
            when COPY =>
                -- internal signals
                next_i          <=  i;
                next_j          <=  j + 1;
                -- outputs
                z_o             <=  a_i;
                wr_o            <=  '1';
                simp_done_o     <=  '0';
            when UPDATE_I =>
                -- internal signals
                next_i          <=  i + 1;
                next_j          <=  j;
                -- outputs
                z_o             <=  a_i;
                wr_o            <=  '0';
                simp_done_o     <=  '0';
            when LAST_COPY =>
                -- internal signals
                next_i          <=  i;
                next_j          <=  j + 1;
                -- outputs
                z_o             <=  a_i;
                wr_o            <=  '1';
                simp_done_o     <=  '0';
            when CHECK_J =>
                -- internal signals
                next_i          <=  i;
                next_j          <=  j;
                -- outputs
                z_o             <=  a_i;
                wr_o            <=  '0';
                simp_done_o     <=  '0';
            when FILL_INVALID =>
                -- internal signals
                next_i          <=  i;
                next_j          <=  j + 1;
                -- outputs
                z_o             <=  INVALID_POS;
                wr_o            <=  '1';
                simp_done_o     <=  '0';
            when DONE =>
                -- internal signals
                next_i          <=  i;
                next_j          <=  j;
                -- outputs
                z_o             <=  a_i;
                wr_o            <=  '0';
                simp_done_o     <=  '1';
            when others =>
                -- internal signals
                next_i          <=  0;
                next_j          <=  0;
                -- outputs
                z_o             <=  a_i;
                wr_o            <=  '0';
                simp_done_o     <=  '0';
        end case;
    end process output_comb;

    output_seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            i           <=  0;
            j           <=  0;
        elsif rising_edge(clk_i) then
            i           <=  next_i;
            j           <=  next_j;
        end if;
    end process output_seq;

end architecture rtl;
