library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.system_params.all;
    use work.matrix_types.all;
    use work.matrix_mult_functions.all;

entity sort is
    port(
        -- control logic
        clk_i:          in  std_logic;
        rst_n_i:        in  std_logic;
        start_i:        in  std_logic;
        -- function arguments in (not latched)
        limit_i:        in  natural;    -- number of elements in array to sort
        base_i:         in  addr;       -- base addr of array to sort
        -- data, addresses and controls to memory
        a_i:            in  pos;        -- element A from array
        a_addr_o:       out addr;       -- address of element A
        b_i:            in  pos;        -- element B from array
        b_addr_o:       out addr;       -- address of element B
        z_o:            out pos;        -- element Z to array
        z_addr_o:       out addr;       -- address of element Z
        wr_o:           out std_logic;  -- write enable for memory
        -- report back once done
        sort_done_o:    out std_logic   -- high when done
    );
end entity sort;

architecture rtl of sort is

    -- element A will be overwritten in the array, so we need to sample it in
    -- order to insert it in the right place later on
    signal  tmp:        pos;
    signal  next_tmp:   pos;

    -- iterators
    signal  i:          natural;
    signal  j:          natural;
    signal  next_i:     natural;
    signal  next_j:     natural;

    -- state machine
    type state_t is (
        IDLE, PREP_INNER, CHECK_INNER, LOOP_INNER,
        LAST_INNER, SAVE_0, SAVE_1, DONE
    );
    signal  state:      state_t;
    signal  next_state: state_t;

begin

    b_addr_o  <=  base_i + to_unsigned(j, address_bits);

    state_comb: process(state, start_i, next_j, next_i, a_i, b_i)
    begin
        case state is
            when IDLE =>
                if start_i = '1' then
                    next_state  <=  PREP_INNER;
                else
                    next_state  <=  IDLE;
                end if;
            when PREP_INNER =>
                next_state  <=  CHECK_INNER;
            when CHECK_INNER =>
                if j /= 0 and b_i > tmp then
                    next_state  <=  LOOP_INNER;
                elsif j = 0 and b_i > tmp then
                    next_state  <=  LAST_INNER;
                else
                    next_state  <=  SAVE_1;
                end if;
            when LOOP_INNER =>
                if next_j /= 0 and a_i > tmp then
                    next_state  <=  LOOP_INNER;
                elsif next_j = 0 and a_i > tmp then
                    next_state  <=  LAST_INNER;
                else
                    next_state  <=  SAVE_1;
                end if;
            when LAST_INNER =>
                next_state  <=  SAVE_0;
            when SAVE_0 =>
                if i /= limit_i - 1 then
                    next_state  <=  PREP_INNER;
                else
                    next_state  <=  DONE;
                end if;
            when SAVE_1 =>
                if i /= limit_i - 1 then
                    next_state  <=  PREP_INNER;
                else
                    next_state  <=  DONE;
                end if;
            when DONE =>
                if start_i = '0' then
                    next_state  <=  IDLE;
                else
                    next_state  <=  DONE;
                end if;
            when others =>
                next_state  <=  IDLE;
        end case;
    end process state_comb;

    state_seq: process(rst_n_i, clk_i)
    begin
        if rst_n_i = '0' then
            state   <=  next_state;
        elsif rising_edge(clk_i) then
            state   <=  next_state;
        end if;
    end process state_seq;

    output_comb: process(state, a_i, b_i, tmp, i, j)
    begin
        case state is
            when IDLE =>
                -- internal signals
                next_tmp    <=  tmp;
                next_i      <=  1;
                next_j      <=  0;
                -- outputs
                a_addr_o    <=  base_i + to_unsigned(i, address_bits);
                z_addr_o    <=  base_i + to_unsigned(j + 1, address_bits);
                z_o         <=  b_i;
                wr_o        <=  '0';
                sort_done_o <=  '0';
            when PREP_INNER =>
                -- internal signals
                next_tmp    <=  a_i;
                next_i      <=  i;
                next_j      <=  i - 1;
                -- outputs
                a_addr_o    <=  base_i + to_unsigned(i, address_bits);
                z_addr_o    <=  base_i + to_unsigned(j + 1, address_bits);
                z_o         <=  b_i;
                wr_o        <=  '0';
                sort_done_o <=  '0';
            when CHECK_INNER =>
                -- internal signals
                next_tmp    <=  tmp;
                next_i      <=  i;
                next_j      <=  j;
                -- outputs
                a_addr_o    <=  base_i + to_unsigned(i, address_bits);
                z_addr_o    <=  base_i + to_unsigned(j + 1, address_bits);
                z_o         <=  b_i;
                wr_o        <=  '0';
                sort_done_o <=  '0';
            when LOOP_INNER =>
                -- internal signals
                next_tmp    <=  tmp;
                next_i      <=  i;
                next_j      <=  j - 1;
                -- outputs
                a_addr_o    <=  base_i + to_unsigned(j - 1, address_bits);
                z_addr_o    <=  base_i + to_unsigned(j + 1, address_bits);
                z_o         <=  b_i;
                wr_o        <=  '1';
                sort_done_o <=  '0';
            when LAST_INNER =>
                -- internal signals
                next_tmp    <=  tmp;
                next_i      <=  i;
                next_j      <=  j;
                -- outputs
                a_addr_o    <=  base_i + to_unsigned(i, address_bits);
                z_addr_o    <=  base_i + to_unsigned(j + 1, address_bits);
                z_o         <=  b_i;
                wr_o        <=  '1';
                sort_done_o <=  '0';
            when SAVE_0 =>
                -- internal signals
                next_tmp    <=  tmp;
                next_i      <=  i + 1;
                next_j      <=  j;
                -- outputs
                a_addr_o    <=  base_i + to_unsigned(i, address_bits);
                z_addr_o    <=  base_i + to_unsigned(j, address_bits);
                z_o         <=  tmp;
                wr_o        <=  '1';
                sort_done_o <=  '0';
            when SAVE_1 =>
                -- internal signals
                next_tmp    <=  tmp;
                next_i      <=  i + 1;
                next_j      <=  j;
                --outputs
                a_addr_o    <=  base_i + to_unsigned(i, address_bits);
                z_addr_o    <=  base_i + to_unsigned(j + 1, address_bits);
                z_o         <=  tmp;
                wr_o        <=  '1';
                sort_done_o <=  '0';
            when DONE =>
                -- internal signals
                next_tmp    <=  tmp;
                next_i      <=  i;
                next_j      <=  j;
                -- outputs
                a_addr_o    <=  base_i + to_unsigned(i, address_bits);
                z_addr_o    <=  base_i + to_unsigned(j + 1, address_bits);
                z_o         <=  tmp;
                wr_o        <=  '0';
                sort_done_o <=  '1';
            when others =>  -- behave like IDLE
                -- internal signals
                next_tmp    <=  tmp;
                next_i      <=  1;
                next_j      <=  0;
                -- outputs
                a_addr_o    <=  base_i + to_unsigned(i, address_bits);
                z_addr_o    <=  base_i + to_unsigned(j + 1, address_bits);
                z_o         <=  b_i;
                wr_o        <=  '0';
                sort_done_o <=  '0';
        end case;
    end process output_comb;

    output_seq: process(rst_n_i, clk_i)
    begin
        if rst_n_i = '0' then
            tmp         <=  (others => '0');
            i           <=  0;
            j           <=  0;
        elsif rising_edge(clk_i) then
            tmp         <=  next_tmp;
            i           <=  next_i;
            j           <=  next_j;
        end if;
    end process output_seq;

end architecture rtl;
