library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.system_params.all;
use work.matrix_types.all;

package matrix_mult_functions is
    function sum(n:n_array) return natural;
    function max(n:n_array) return natural;
    function get_block_limits return n_matrix;
    function get_Htr_block_base return addr_array;
    function get_Qtr_block_base return addr_matrix;
end package matrix_mult_functions;

package body matrix_mult_functions is
    function sum(n:n_array) return natural is
        variable    sum:    natural;
    begin
        sum :=  0;
        for i in 0 to n'high loop
            sum :=  sum + n(i);
        end loop;
        return sum;
    end function sum;

    function max(n:n_array) return natural is
        variable    max:    natural;
    begin
        max :=  0;
        for i in 0 to n'high loop
            if max < n(i) then
                max :=  n(i);
            end if;
        end loop;
        return max;
    end function max;

    function get_block_limits return n_matrix is
        -- Mtr_matrix: count of elements in each block
        variable    Mtr_matrix: n_matrix(0 to N0-1, 0 to N0-1);
        -- M_limits: as above decreased by 1 (counter limits)
        variable    M_limits:   n_matrix(0 to N0-1, 0 to N0-1);
        variable    row:        n_array(0 to N0-1);
        variable    tmp:        natural;
    begin
        Mtr_matrix  :=  (others => (others => 0));
        -- invert M to find last row of Mtr matrix
        for i in 0 to N0-1 loop
            row(i)  :=  M(N0-1-i);
        end loop;
        -- rotate row right to find first row of Mtr matrix
        -- (not fundamental, just to keep things the usual order)
        tmp         :=  row(N0-1);
        for i in N0-1 downto 1 loop
            row(i)  :=  row(i-1);
        end loop;
        row(0)   :=  tmp;
        for i in 0 to N0-1 loop
            -- copy row
            for j in 0 to N0-1 loop
                Mtr_matrix(i, j)  :=  row(j);
            end loop;
            -- get new row
            tmp         :=  row(N0-1);
            for i in N0-1 downto 1 loop
                row(i)  :=  row(i-1);
            end loop;
            row(0)   :=  tmp;
        end loop;
        -- copy each element of the Mtr matrix decreasing it by 1
        for i in 0 to N0-1 loop
            for j in 0 to N0-1 loop
                M_limits(i, j)  :=  Mtr_matrix(i, j) - 1;
            end loop;
        end loop;
        return M_limits;
    end function get_block_limits;

    function get_Htr_block_base return addr_array is
        variable    Htr_block_base: addr_array(0 to N0-1);
    begin
        Htr_block_base  :=  (others => (others => '0'));
        for i in 0 to N0-1 loop
            Htr_block_base(i)   :=  to_unsigned(i*DV, address_bits);
        end loop;
        return Htr_block_base;
    end function get_Htr_block_base;

    function get_Qtr_block_base return addr_matrix is
        variable    Qtr_block_base: addr_matrix(0 to N0-1, 0 to N0-1);
        variable    Mtr:            addr_matrix(0 to N0-1, 0 to N0-1);
        variable    row:            n_array(0 to N0-1);
        variable    tmp:            natural;
    begin
        Qtr_block_base  :=  (others => (others => (others => '0')));
        Mtr             :=  (others => (others => (others => '0')));
        -- invert M to find last row of Mtr matrix
        for i in 0 to N0-1 loop
            row(i)  :=  M(N0-1-i);
        end loop;
        -- rotate row right to find first row of Mtr matrix
        -- (not fundamental, just to keep things the usual order)
        tmp         :=  row(N0-1);
        for i in N0-1 downto 1 loop
            row(i)  :=  row(i-1);
        end loop;
        row(0)   :=  tmp;
        for i in 0 to N0-1 loop
            -- copy row
            for j in 0 to N0-1 loop
                Mtr(i, j)   :=  to_unsigned(row(j), address_bits);
            end loop;
            -- get new row
            tmp         :=  row(N0-1);
            for i in N0-1 downto 1 loop
                row(i)  :=  row(i-1);
            end loop;
            row(0)   :=  tmp;
        end loop;
        -- Get block base as previous base plus limit of previous block
        Qtr_block_base(0, 0)    :=  (others => '0');
        for j in 1 to N0-1 loop
            Qtr_block_base(0, j)    :=  Qtr_block_base(0, j-1) + Mtr(0, j-1);
        end loop;
        for i in 1 to N0-1 loop
            Qtr_block_base(i, 0)    :=  Qtr_block_base(i-1, N0-1) + Mtr(i-1, N0-1);
            for j in 1 to N0-1 loop
                Qtr_block_base(i, j)    :=  Qtr_block_base(i, j-1) + Mtr(i, j-1);
            end loop;
        end loop;
        return Qtr_block_base;
    end function get_Qtr_block_base;
end package body matrix_mult_functions;
