library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.system_params.all;
    use work.matrix_types.all;
    use work.matrix_mult_functions.all;

entity top is
    port(
        -- control signals
        clk_i:      in  std_logic;
        rst_n_i:    in  std_logic;
        start_i:    in  std_logic;
        -- to memory
        a_i:        in  pos;
        a_addr_o:   out addr;
        b_i:        in  pos;
        b_addr_o:   out addr;
        z_o:        out pos;
        z_addr_o:   out addr;
        wr_o:       out std_logic;
        -- report back when done
        failure_o:  out std_logic;
        done_o:     out std_logic
    );
end entity top;

architecture rtl of top is

    component key_reconstruction is
        port (
            -- control signals
            clk_i:          in  std_logic;
            rst_n_i:        in  std_logic;
            start_i:        in  std_logic;
            -- data, addresses and controls to memory
            a_i:            in  pos;
            a_addr_o:       out addr;
            b_i:            in  pos;
            b_addr_o:       out addr;
            z_o:            out pos;
            z_addr_o:       out addr;
            wr_o:           out std_logic;
            -- report back once done
            key_rec_done_o: out std_logic
        );
    end component key_reconstruction;
    signal  kr_start:   std_logic;
    signal  kr_a_addr:  addr;
    signal  kr_b_addr:  addr;
    signal  kr_z:       pos;
    signal  kr_z_addr:  addr;
    signal  kr_wr:      std_logic;
    signal  kr_done:    std_logic;

    component vector_by_circulant is
        port(
            -- control signals
            clk_i:          in  std_logic;
            rst_n_i:        in  std_logic;
            start_i:        in  std_logic;
            binary_i:       in  std_logic;
            -- function arguments in (not latched)
            a_base_i:       in  addr;
            b_base_i:       in  addr;
            b_weight:       in  integer;
            z_base_i:       in  addr;
            op:             in  std_logic;  -- '0' -> z=aB; '1' -> z=Ba 
            -- data, addresses and controls to memory
            a_i:            in  pos;
            a_addr_o:       out addr;
            b_i:            in  pos;
            b_addr_o:       out addr;
            z_o:            out pos;
            z_addr_o:       out addr;
            wr_o:           out std_logic;
            done_o:         out std_logic
        );
    end component vector_by_circulant;
    signal  vc_start:       std_logic;
    signal  vc_binary:      std_logic;
    signal  vc_vec_base:    addr;
    signal  vc_blk_base:    addr;
    signal  vc_blk_weight:  natural;
    signal  vc_res_base:    addr;
    signal  vc_op:          std_logic;
    signal  vc_a_addr:      addr;
    signal  vc_b_addr:      addr;
    signal  vc_z:           pos;
    signal  vc_z_addr:      addr;
    signal  vc_wr:          std_logic;
    signal  vc_done:        std_logic;

    component x_by_Ltr is
        port (
            -- control signals
            clk_i:          in  std_logic;
            rst_n_i:        in  std_logic;
            start_i:        in  std_logic;
            -- controls to multiplication core
            vector_base_o:  out addr;
            block_base_o:   out addr;
            result_base_o:  out addr;
            start_mul_o:    out std_logic;
            mul_done_i:     in  std_logic;
            -- report back
            done_o:         out std_logic
        );
    end component x_by_Ltr;
    signal  xl_start:       std_logic;
    signal  xl_vec_base:    addr;
    signal  xl_blk_base:    addr;
    signal  xl_res_base:    addr;
    signal  xl_start_mul:   std_logic;
    signal  xl_done:        std_logic;

    component Htr_by_str is
        port (
            -- control signals
            clk_i:          in  std_logic;
            rst_n_i:        in  std_logic;
            start_i:        in  std_logic;
            -- controls to multiplication core
            vector_base_o:  out addr;
            block_base_o:   out addr;
            result_base_o:  out addr;
            start_mul_o:    out std_logic;
            mul_done_i:     in  std_logic;
            -- report back
            done_o:         out std_logic
        );
    end component Htr_by_str;
    signal  hs_start:       std_logic;
    signal  hs_vec_base:    addr;
    signal  hs_blk_base:    addr;
    signal  hs_res_base:    addr;
    signal  hs_start_mul:   std_logic;
    signal  hs_done:        std_logic;

    component Qtr_by_sigmatr is
        port (
            -- control signals
            clk_i:          in  std_logic;
            rst_n_i:        in  std_logic;
            start_i:        in  std_logic;
            -- controls to multiplication core
            vector_base_o:  out addr;
            block_base_o:   out addr;
            block_weight_o: out natural;
            result_base_o:  out addr;
            start_mul_o:    out std_logic;
            mul_done_i:     in  std_logic;
            -- report back
            done_o:         out std_logic
        );
    end component Qtr_by_sigmatr;
    signal  qs_start:       std_logic;
    signal  qs_vec_base:    addr;
    signal  qs_blk_base:    addr;
    signal  qs_blk_weight:  natural;
    signal  qs_res_base:    addr;
    signal  qs_start_mul:   std_logic;
    signal  qs_done:        std_logic;

    component find_max is
        port(
            -- control signals
            clk_i:      in  std_logic;
            rst_n_i:    in  std_logic;
            start_i:    in  std_logic;
            -- to memory
            a_i:        in  pos;
            a_addr_o:   out addr;
            b_i:        in  pos;
            b_addr_o:   out addr;
            z_o:        out pos;
            z_addr_o:   out addr;
            wr_o:       out std_logic;
            -- report back when done
            max_idx_o:  out n_array(15 downto 0);
            done_o:     out std_logic
        );
    end component find_max;
    signal  fm_start:       std_logic;
    signal  fm_a_addr:      addr;
    signal  fm_b_addr:      addr;
    signal  fm_z:           pos;
    signal  fm_z_addr:      addr;
    signal  fm_wr:          std_logic;
    signal  fm_max_idx:     n_array(15 downto 0);
    signal  fm_done:        std_logic;
    signal  max_idx_c:      n_array(15 downto 0);
    signal  max_idx_r:      n_array(15 downto 0);

    component vector_plus_rows is
        port (
            -- control signals
            clk_i:      in  std_logic;
            rst_n_i:    in  std_logic;
            start_i:    in  std_logic;
            -- row indexes
            row_idx_i:  in  n_array(15 downto 0);
            -- to memory
            a_i:        in  pos;
            a_addr_o:   out addr;
            b_i:        in  pos;
            b_addr_o:   out addr;
            z_o:        out pos;
            z_addr_o:   out addr;
            wr_o:       out std_logic;
            -- report back when done
            done_o:     out std_logic
        );
    end component vector_plus_rows;
    signal  vr_start:       std_logic;
    signal  vr_a_addr:      addr;
    signal  vr_b_addr:      addr;
    signal  vr_z:           pos;
    signal  vr_z_addr:      addr;
    signal  vr_wr:          std_logic;
    signal  vr_done:        std_logic;

    component e_by_Htr is
        port (
            -- control signals
            clk_i:          in  std_logic;
            rst_n_i:        in  std_logic;
            start_i:        in  std_logic;
            -- controls to multiplication core
            vector_base_o:  out addr;
            block_base_o:   out addr;
            result_base_o:  out addr;
            start_mul_o:    out std_logic;
            mul_done_i:     in  std_logic;
            -- report back
            done_o:         out std_logic
        );
    end component e_by_Htr;
    signal  eh_start:       std_logic;
    signal  eh_vec_base:    addr;
    signal  eh_blk_base:    addr;
    signal  eh_res_base:    addr;
    signal  eh_start_mul:   std_logic;
    signal  eh_done:        std_logic;

    component null_syndrome is
        port(
            -- control signals
            clk_i:      in  std_logic;
            rst_n_i:    in  std_logic;
            start_i:    in  std_logic;
            -- to memory
            a_i:        in  pos;
            a_addr_o:   out addr;
            b_i:        in  pos;
            b_addr_o:   out addr;
            z_o:        out pos;
            z_addr_o:   out addr;
            wr_o:       out std_logic;
            -- report back when done
            null_syn_o: out std_logic;
            done_o:     out std_logic
        );
    end component null_syndrome;
    signal  ns_start:       std_logic;
    signal  ns_a_addr:      addr;
    signal  ns_b_addr:      addr;
    signal  ns_z:           pos;
    signal  ns_z_addr:      addr;
    signal  ns_wr:          std_logic;
    signal  ns_null_syn:    std_logic;
    signal  ns_done:        std_logic;

    component clear_temp is
        port(
            -- control signals
            clk_i:      in  std_logic;
            rst_n_i:    in  std_logic;
            start_i:    in  std_logic;
            -- to memory
            a_i:        in  pos;
            a_addr_o:   out addr;
            b_i:        in  pos;
            b_addr_o:   out addr;
            z_o:        out pos;
            z_addr_o:   out addr;
            wr_o:       out std_logic;
            -- report back when done
            done_o:     out std_logic
        );
    end component clear_temp;
    signal  ct_start:       std_logic;
    signal  ct_a_addr:      addr;
    signal  ct_b_addr:      addr;
    signal  ct_z:           pos;
    signal  ct_z_addr:      addr;
    signal  ct_wr:          std_logic;
    signal  ct_null_syn:    std_logic;
    signal  ct_done:        std_logic;

    component comp_message is
        port(
            -- control signals
            clk_i:      in  std_logic;
            rst_n_i:    in  std_logic;
            start_i:    in  std_logic;
            -- to memory
            a_i:        in  pos;
            a_addr_o:   out addr;
            b_i:        in  pos;
            b_addr_o:   out addr;
            z_o:        out pos;
            z_addr_o:   out addr;
            wr_o:       out std_logic;
            -- report back when done
            done_o:     out std_logic
        );
    end component comp_message;
    signal  cm_start:       std_logic;
    signal  cm_a_addr:      addr;
    signal  cm_b_addr:      addr;
    signal  cm_z:           pos;
    signal  cm_z_addr:      addr;
    signal  cm_wr:          std_logic;
    signal  cm_done:        std_logic;

    type state_t is (
        IDLE,
        COMPUTE_LTR,
        COMPUTE_INITIAL_SYNDROME,
        COMPUTE_SIGMA,
        COMPUTE_R,
        FIND_B,
        COMPUTE_ERROR,
        COMPUTE_SYNDROME,
        CHECK_SYNDROME,
        CLEAR_TEMP_AND_LOOP,
        COMPUTE_MESSAGE,
        CLEAR_TEMP_AND_RETURN,
        DONE
    );
    signal  state:      state_t;
    signal  next_state: state_t;

    signal  l_c:    integer;
    signal  l_r:    integer;

begin

    KR: key_reconstruction
        port map (
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  kr_start,
            a_i             =>  a_i,
            a_addr_o        =>  kr_a_addr,
            b_i             =>  b_i,
            b_addr_o        =>  kr_b_addr,
            z_o             =>  kr_z,
            z_addr_o        =>  kr_z_addr,
            wr_o            =>  kr_wr,
            key_rec_done_o  =>  kr_done
        );

    VC: vector_by_circulant
        port map (
            clk_i       =>  clk_i,
            rst_n_i     =>  rst_n_i,
            start_i     =>  vc_start,
            binary_i    =>  vc_binary,
            a_base_i    =>  vc_vec_base,
            b_base_i    =>  vc_blk_base,
            b_weight    =>  vc_blk_weight,
            z_base_i    =>  vc_res_base,
            op          =>  vc_op,
            a_i         =>  a_i,
            a_addr_o    =>  vc_a_addr,
            b_i         =>  b_i,          
            b_addr_o    =>  vc_b_addr,
            z_o         =>  vc_z,
            z_addr_o    =>  vc_z_addr,
            wr_o        =>  vc_wr,
            done_o      =>  vc_done
        );

    XL: x_by_Ltr
        port map (
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  xl_start,
            vector_base_o   =>  xl_vec_base,
            block_base_o    =>  xl_blk_base,
            result_base_o   =>  xl_res_base,
            start_mul_o     =>  xl_start_mul,
            mul_done_i      =>  vc_done,
            done_o          =>  xl_done
        );

    HS: Htr_by_str
        port map (
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  hs_start,
            vector_base_o   =>  hs_vec_base,
            block_base_o    =>  hs_blk_base,
            result_base_o   =>  hs_res_base,
            start_mul_o     =>  hs_start_mul,
            mul_done_i      =>  vc_done,
            done_o          =>  hs_done
        );

    QS: Qtr_by_sigmatr
        port map (
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  qs_start,
            vector_base_o   =>  qs_vec_base,
            block_base_o    =>  qs_blk_base,
            block_weight_o  =>  qs_blk_weight,
            result_base_o   =>  qs_res_base,
            start_mul_o     =>  qs_start_mul,
            mul_done_i      =>  vc_done,
            done_o          =>  qs_done
        );

    FM: find_max
        port map (
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  fm_start,
            a_i             =>  a_i,
            a_addr_o        =>  fm_a_addr,
            b_i             =>  b_i,
            b_addr_o        =>  fm_b_addr,
            z_o             =>  fm_z,
            z_addr_o        =>  fm_z_addr,
            wr_o            =>  fm_wr,
            max_idx_o       =>  fm_max_idx,
            done_o          =>  fm_done
        );

    VR: vector_plus_rows
        port map (
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  vr_start,
            row_idx_i       =>  max_idx_r,
            a_i             =>  a_i,
            a_addr_o        =>  vr_a_addr,
            b_i             =>  b_i,
            b_addr_o        =>  vr_b_addr,
            z_o             =>  vr_z,
            z_addr_o        =>  vr_z_addr,
            wr_o            =>  vr_wr,
            done_o          =>  vr_done
        );

    EH: e_by_Htr
        port map (
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  eh_start,
            vector_base_o   =>  eh_vec_base,
            block_base_o    =>  eh_blk_base,
            result_base_o   =>  eh_res_base,
            start_mul_o     =>  eh_start_mul,
            mul_done_i      =>  vc_done,
            done_o          =>  eh_done
        );

    NS: null_syndrome
        port map (
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  ns_start,
            a_i             =>  a_i,
            a_addr_o        =>  ns_a_addr,
            b_i             =>  b_i,
            b_addr_o        =>  ns_b_addr,
            z_o             =>  ns_z,
            z_addr_o        =>  ns_z_addr,
            wr_o            =>  ns_wr,
            null_syn_o      =>  ns_null_syn,
            done_o          =>  ns_done
        );

    CT: clear_temp
        port map (
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  ct_start,
            a_i             =>  a_i,
            a_addr_o        =>  ct_a_addr,
            b_i             =>  b_i,
            b_addr_o        =>  ct_b_addr,
            z_o             =>  ct_z,
            z_addr_o        =>  ct_z_addr,
            wr_o            =>  ct_wr,
            done_o          =>  ct_done
        );

    CM: comp_message
        port map (
            clk_i           =>  clk_i,
            rst_n_i         =>  rst_n_i,
            start_i         =>  cm_start,
            a_i             =>  a_i,
            a_addr_o        =>  cm_a_addr,
            b_i             =>  b_i,
            b_addr_o        =>  cm_b_addr,
            z_o             =>  cm_z,
            z_addr_o        =>  cm_z_addr,
            wr_o            =>  cm_wr,
            done_o          =>  cm_done
        );

    seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            state       <=  IDLE;
            max_idx_r   <=  (others => 0);
            l_r         <=  0;
        elsif rising_edge(clk_i) then
            state       <=  next_state;
            max_idx_r   <=  max_idx_c;
            l_r         <=  l_c;
        end if;
    end process seq;
    -- TODO: put everything in sensitivity list
    comb: process(all)
--state, start_i,
--        kr_a_addr, kr_b_addr, kr_z, kr_z_addr, kr_wr, kr_done,
--        vc_a_addr, vc_b_addr, vc_z, vc_z_addr, vc_wr,
--        xl_start_mul, xl_vec_base, xl_blk_base, xl_res_base, xl_done,
--        hs_start_mul, hs_vec_base, hs_blk_base, hs_res_base, hs_done,
--        qs_start_mul, qs_vec_base, qs_blk_base, qs_res_base, qs_done,
--        fm_a_addr, fm_b_addr, fm_z, fm_z_addr, fm_wr, fm_max_idx, fm_done,
--        vr_a_addr, vr_b_addr, vr_z, vr_z_addr, vr_wr, vr_done,
--        eh_start_mul, eh_vec_base, eh_blk_base, eh_res_base, eh_done)
    begin
        next_state      <=  state;
        a_addr_o        <=  to_unsigned(0, address_bits);
        b_addr_o        <=  to_unsigned(0, address_bits);
        z_o             <=  to_unsigned(0, position_bits);
        z_addr_o        <=  to_unsigned(0, address_bits);
        wr_o            <=  '0';
        failure_o       <=  '0';
        done_o          <=  '0';
        kr_start        <=  '0';
        vc_start        <=  '0';
        vc_binary       <=  '0';
        vc_vec_base     <=  to_unsigned(0, address_bits);
        vc_blk_base     <=  to_unsigned(0, address_bits);
        vc_blk_weight   <=  0;
        vc_res_base     <=  to_unsigned(0, address_bits);
        vc_op           <=  '0';
        xl_start        <=  '0';
        hs_start        <=  '0';
        qs_start        <=  '0';
        fm_start        <=  '0';
        max_idx_c       <=  max_idx_r;
        vr_start        <=  '0';
        eh_start        <=  '0';
        ns_start        <=  '0';
        ct_start        <=  '0';
        cm_start        <=  '0';
        l_c             <=  l_r;
        case state is
            when IDLE =>
                l_c <=  0;
                if start_i = '1' then
                    next_state  <=  COMPUTE_LTR;
                end if;
            when COMPUTE_LTR =>
                a_addr_o    <=  kr_a_addr;
                b_addr_o    <=  kr_b_addr;
                z_o         <=  kr_z;
                z_addr_o    <=  kr_z_addr;
                wr_o        <=  kr_wr;
                kr_start    <=  '1';
                if kr_done = '1' then
                    next_state  <=  COMPUTE_INITIAL_SYNDROME;
                end if;
            when COMPUTE_INITIAL_SYNDROME =>
                a_addr_o        <=  vc_a_addr;
                b_addr_o        <=  vc_b_addr;
                z_o             <=  vc_z;
                z_addr_o        <=  vc_z_addr;
                wr_o            <=  vc_wr;
                vc_start        <=  xl_start_mul;
                vc_binary       <=  '1';
                vc_vec_base     <=  xl_vec_base;
                vc_blk_base     <=  xl_blk_base;
                vc_blk_weight   <=  sum(M)*DV;
                vc_res_base     <=  xl_res_base;
                vc_op           <=  '0';
                xl_start        <=  '1';
                if xl_done = '1' then
                    next_state  <=  COMPUTE_SIGMA;
                end if;
            when COMPUTE_SIGMA =>
                a_addr_o        <=  vc_a_addr;
                b_addr_o        <=  vc_b_addr;
                z_o             <=  vc_z;
                z_addr_o        <=  vc_z_addr;
                wr_o            <=  vc_wr;
                vc_start        <=  hs_start_mul;
                vc_binary       <=  '0';
                vc_vec_base     <=  hs_vec_base;
                vc_blk_base     <=  hs_blk_base;
                vc_blk_weight   <=  DV;
                vc_res_base     <=  hs_res_base;
                vc_op           <=  '1';
                hs_start        <=  '1';
                if hs_done = '1' then
                    next_state  <=  COMPUTE_R;
                end if;
            when COMPUTE_R =>
                a_addr_o        <=  vc_a_addr;
                b_addr_o        <=  vc_b_addr;
                z_o             <=  vc_z;
                z_addr_o        <=  vc_z_addr;
                wr_o            <=  vc_wr;
                vc_start        <=  qs_start_mul;
                vc_binary       <=  '0';
                vc_vec_base     <=  qs_vec_base;
                vc_blk_base     <=  qs_blk_base;
                vc_blk_weight   <=  qs_blk_weight + 1;
                vc_res_base     <=  qs_res_base;
                vc_op           <=  '1';
                qs_start        <=  '1';
                if qs_done = '1' then
                    next_state  <=  FIND_B;
                end if;
            when FIND_B =>
                a_addr_o        <=  fm_a_addr;
                b_addr_o        <=  fm_b_addr;
                z_o             <=  fm_z;
                z_addr_o        <=  fm_z_addr;
                wr_o            <=  fm_wr;
                fm_start        <=  '1';
                if fm_done = '1' then
                    max_idx_c   <=  fm_max_idx;
                    next_state  <=  COMPUTE_ERROR;
                end if;
            when COMPUTE_ERROR =>
                a_addr_o        <=  vr_a_addr;
                b_addr_o        <=  vr_b_addr;
                z_o             <=  vr_z;
                z_addr_o        <=  vr_z_addr;
                wr_o            <=  vr_wr;
                vr_start        <=  '1';
                if vr_done = '1' then
                    next_state  <=  COMPUTE_SYNDROME;
                end if;
            when COMPUTE_SYNDROME =>
                a_addr_o        <=  vc_a_addr;
                b_addr_o        <=  vc_b_addr;
                z_o             <=  vc_z;
                z_addr_o        <=  vc_z_addr;
                wr_o            <=  vc_wr;
                vc_start        <=  eh_start_mul;
                vc_binary       <=  '1';
                vc_vec_base     <=  eh_vec_base;
                vc_blk_base     <=  eh_blk_base;
                vc_blk_weight   <=  sum(M)*DV;
                vc_res_base     <=  eh_res_base;
                vc_op           <=  '0';
                eh_start        <=  '1';
                if eh_done = '1' then
                    next_state  <=  CHECK_SYNDROME;
                end if;
            when CHECK_SYNDROME =>
                a_addr_o        <=  ns_a_addr;
                b_addr_o        <=  ns_b_addr;
                z_o             <=  ns_z;
                z_addr_o        <=  ns_z_addr;
                wr_o            <=  ns_wr;
                ns_start        <=  '1';
                if ns_done = '1' then
                    if ns_null_syn = '1' or l_r >= 20 then
                        next_state  <=  COMPUTE_MESSAGE;
                    else
                        next_state  <=  CLEAR_TEMP_AND_LOOP;
                    end if;
                end if;
            when CLEAR_TEMP_AND_LOOP =>
                a_addr_o        <=  ct_a_addr;
                b_addr_o        <=  ct_b_addr;
                z_o             <=  ct_z;
                z_addr_o        <=  ct_z_addr;
                wr_o            <=  ct_wr;
                ct_start        <=  '1';
                if ct_done = '1' then
                    l_c         <=  l_r + 1;
                    next_state  <=  COMPUTE_SIGMA;
                end if;
            when COMPUTE_MESSAGE =>
                a_addr_o        <=  cm_a_addr;
                b_addr_o        <=  cm_b_addr;
                z_o             <=  cm_z;
                z_addr_o        <=  cm_z_addr;
                wr_o            <=  cm_wr;
                cm_start        <=  '1';
                if cm_done = '1' then
                    next_state  <=  CLEAR_TEMP_AND_RETURN;
                end if;
            when CLEAR_TEMP_AND_RETURN =>
                a_addr_o        <=  ct_a_addr;
                b_addr_o        <=  ct_b_addr;
                z_o             <=  ct_z;
                z_addr_o        <=  ct_z_addr;
                wr_o            <=  ct_wr;
                ct_start        <=  '1';
                if ct_done = '1' then
                    next_state  <=  DONE;
                end if;
            when DONE =>
                done_o  <=  '1';
                if l_r >= 20 then
                    failure_o   <=  '1';
                end if;
                if start_i = '0' then
                    next_state  <=  IDLE;
                end if;
            when others =>
                null;
        end case;
    end process comb;

end architecture rtl;

