library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.system_params.all;
    use work.matrix_types.all;
    use work.matrix_mult_functions.all;
    use work.memory_map.all;

entity Qtr_by_sigmatr is
    port (
        -- control signals
        clk_i:          in  std_logic;
        rst_n_i:        in  std_logic;
        start_i:        in  std_logic;
        -- controls to multiplication core
        vector_base_o:  out addr;
        block_base_o:   out addr;
        block_weight_o: out natural;
        result_base_o:  out addr;
        start_mul_o:    out std_logic;
        mul_done_i:     in  std_logic;
        -- report back
        done_o:         out std_logic
    );
end entity Qtr_by_sigmatr;

architecture rtl of Qtr_by_sigmatr is

    type state_t is (IDLE, BUSY, DONE);
    signal  state:      state_t;
    signal  next_state: state_t;

    signal  block_row_c, block_row_r:       natural;
    signal  block_col_c, block_col_r:       natural;

    signal  Qtr_block_base_map:
        addr_matrix(N0-1 downto 0, N0-1 downto 0)   :=  get_Qtr_block_base;
    signal  Qtr_block_limit_map:
        n_matrix(N0-1 downto 0, N0-1 downto 0)      :=  get_block_limits;

    signal  vector_base_c, vector_base_r:   addr;
    signal  block_base_c, block_base_r:     addr;
    signal  result_base_c, result_base_r:   addr;

begin

    vector_base_o   <=  vector_base_r;
    block_base_o    <=  block_base_r;
    block_weight_o  <=  Qtr_block_limit_map(block_row_r, block_col_r);
    result_base_o   <=  result_base_r;

    comb: process(state, start_i,
        block_base_r, vector_base_r, result_base_r,
        block_row_r, block_col_r,
        mul_done_i)
    begin
        start_mul_o     <=  '0';
        next_state      <=  state;
        block_row_c     <=  block_row_r;
        block_col_c     <=  block_col_r;
        block_base_c    <=  block_base_r;
        vector_base_c   <=  vector_base_r;
        result_base_c   <=  result_base_r;
        done_o          <=  '0';
        case state is
            when IDLE =>
                block_row_c     <=  0;
                block_col_c     <=  0;
                block_base_c    <=  QTR_BASE_ADDR;
                vector_base_c   <=  SIGMA_BASE_ADDR;
                result_base_c   <=  R_BASE_ADDR;
                if start_i = '1' then
                    next_state  <=  BUSY;
                end if;
            when BUSY =>
                start_mul_o <=  '1';
                if mul_done_i = '1' then
                    start_mul_o     <=  '0';
                    if block_col_r /= N0-1 then
                        block_col_c     <=  block_col_r + 1;
                        block_base_c    <=  QTR_BASE_ADDR +  -- cont.
                            Qtr_block_base_map(block_row_r, block_col_r);
                        vector_base_c   <=  vector_base_r + P;
                        result_base_c   <=  result_base_r;
                    else
                        if block_row_r /= N0-1 then
                            block_row_c     <=  block_row_r + 1;
                            block_col_c     <=  0;
                            block_base_c    <=  QTR_BASE_ADDR + -- cont.
                                Qtr_block_base_map(block_row_r + 1, -- cont.
                                    block_col_r);
                            vector_base_c   <=  SIGMA_BASE_ADDR;
                            result_base_c   <=  result_base_r + P;
                        else
                            next_state  <=  DONE;
                        end if;
                    end if;
                end if;
            when DONE =>
                done_o  <=  '1';
                if start_i = '0' then
                    next_state  <=  IDLE;
                end if;
            when others =>
                next_state  <=  IDLE;
        end case;
    end process comb;

    seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            state           <=  IDLE;
            block_row_r     <=  0;
            block_col_r     <=  0;
            block_base_r    <=  HTR_BASE_ADDR;
            vector_base_r   <=  S_BASE_ADDR; 
            result_base_r   <=  SIGMA_BASE_ADDR;
        elsif rising_edge(clk_i) then
            state           <=  next_state;
            block_base_r    <=  block_base_c;
            vector_base_r   <=  vector_base_c;
            result_base_r   <=  result_base_c;
            block_row_r     <=  block_row_c;
            block_col_r     <=  block_col_c;
        end if;
    end process seq;

end architecture rtl;

