library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.system_params.all;
    use work.matrix_types.all;

entity vector_by_circulant is
    port(
        -- control signals
        clk_i:          in  std_logic;
        rst_n_i:        in  std_logic;
        start_i:        in  std_logic;
        binary_i:       in  std_logic;
        -- function arguments in (not latched)
        a_base_i:       in  addr;
        b_base_i:       in  addr;
        b_weight:       in  natural;
        z_base_i:       in  addr;
        op:             in  std_logic;  -- '0' -> z=aB; '1' -> z=Ba 
        -- data, addresses and controls to memory
        a_i:            in  pos;
        a_addr_o:       out addr;
        b_i:            in  pos;
        b_addr_o:       out addr;
        z_o:            out pos;
        z_addr_o:       out addr;
        wr_o:           out std_logic;
        done_o:         out std_logic
    );
end entity vector_by_circulant;

architecture arch of vector_by_circulant is

    type state_t is (IDLE, GET_M, INNER_LOOP, DONE);
    signal  state:      state_t;
    signal  next_state: state_t;

    signal  m_index_c, m_index_r:   natural;
    signal  m_c, m_r:   pos;
    signal  n_c, n_r:   addr;
    signal  i_c, i_r:   addr;

begin

    comb: process(state, start_i, binary_i,
        m_index_r, m_c, m_r, n_c, n_r, i_c, i_r,
        op, b_weight, a_base_i, b_base_i, z_base_i,
        a_i, b_i)
    begin
        next_state  <=  state;
        wr_o        <=  '0';
        done_o      <=  '0';
        m_index_c   <=  m_index_r;
        a_addr_o    <=  to_unsigned(0, address_bits);
        b_addr_o    <=  to_unsigned(0, address_bits);
        z_o         <=  to_unsigned(0, position_bits);
        z_addr_o    <=  to_unsigned(0, address_bits);
        m_c         <=  m_r;
        n_c         <=  n_r;
        i_c         <=  i_r;
        case state is
            when IDLE =>
                m_index_c   <=  0;
                m_c         <=  to_unsigned(0, position_bits);
                n_c         <=  to_unsigned(0, address_bits);
                i_c         <=  to_unsigned(0, address_bits);
                if start_i = '1' then
                    next_state  <=  GET_M;
                end if;
            when GET_M =>
                m_index_c   <=  m_index_r + 1;
                b_addr_o    <=  b_base_i + m_index_r;
                m_c         <=  b_i;
                n_c         <=  to_unsigned(0, address_bits);
                if op = '0' then
                    i_c     <=  to_unsigned(to_integer(m_c), address_bits);
                else
                    i_c     <=  to_unsigned(to_integer(P-m_c-1),  -- cont.
                                            address_bits);
                end if;
                if b_i = P then
                    next_state  <=  DONE;
                else
                    next_state  <=  INNER_LOOP;
                end if;
            when INNER_LOOP =>
                wr_o    <=  '1';
                n_c     <=  n_r + 1;
                if i_r = P - 1 then
                    i_c <=  to_unsigned(0, address_bits);
                else
                    i_c <=  i_r + 1;
                end if;
                a_addr_o    <=  a_base_i + n_c;
                b_addr_o    <=  z_base_i + i_c;
                z_addr_o    <=  z_base_i + i_c;
                if binary_i = '0' then
                    z_o     <=  a_i + b_i;
                else
                    z_o     <=  (0 => a_i(0) xor b_i(0), others => '0');
                end if;
                if n_r = P-1 then
                    if m_index_r = b_weight then
                        next_state  <=  DONE;
                    else
                        next_state  <=  GET_M;
                    end if;
                end if;
            when DONE =>
                done_o  <=  '1';
                if start_i = '0' then
                    next_state  <=  IDLE;
                end if;
            when others =>
                next_state  <=  IDLE;
        end case;
    end process comb;

    seq: process(rst_n_i, clk_i)
    begin
        if rst_n_i = '0' then
            state       <=  IDLE;
            m_index_r   <=  0;
            m_r         <=  to_unsigned(0, position_bits);
            n_r         <=  to_unsigned(0, address_bits);
            i_r         <=  to_unsigned(0, address_bits);
        elsif rising_edge(clk_i) then
            state       <=  next_state;
            m_index_r   <=  m_index_c;
            m_r         <=  m_c;
            n_r         <=  n_c;
            i_r         <=  i_c;
        end if;
    end process seq;

end architecture;

