library IEEE;
    use IEEE.std_logic_1164.all;
    use IEEE.numeric_std.all;
library work;
    use work.system_params.all;
    use work.matrix_types.all;
    use work.memory_map.all;
    use work.matrix_mult_functions.all;

entity vector_plus_rows is
    port (
        -- control signals
        clk_i:      in  std_logic;
        rst_n_i:    in  std_logic;
        start_i:    in  std_logic;
        -- row indexes
        row_idx_i:  in  n_array(15 downto 0);
        -- to memory
        a_i:        in  pos;
        a_addr_o:   out addr;
        b_i:        in  pos;
        b_addr_o:   out addr;
        z_o:        out pos;
        z_addr_o:   out addr;
        wr_o:       out std_logic;
        -- report back when done
        done_o:     out std_logic
    );
end entity vector_plus_rows;

architecture rtl of vector_plus_rows is

    type state_t is (IDLE, RELATIVE_ROW, SUM_ROW, DONE);
    signal  state:      state_t;
    signal  next_state: state_t;

    signal  cur_idx_c:  natural;
    signal  cur_idx_r:  natural;

    signal  block_row_c:    natural;
    signal  block_row_r:    natural;
    signal  block_col_c:    natural;
    signal  block_col_r:    natural;
    signal  rel_row_c:      natural;
    signal  rel_row_r:      natural;

    signal  i_c:    natural;
    signal  i_r:    natural;

    signal  Qtr_block_base:     addr;
    signal  Qtr_block_count:    natural;

    constant    Qtr_block_limit_map:    n_matrix:=      get_block_limits;
    constant    Qtr_block_base_map:     addr_matrix:=   get_Qtr_block_base;

    constant    INVALID:    natural:=   N0*P;

begin

    Qtr_block_count <=  Qtr_block_limit_map(block_row_r, block_col_r);
    Qtr_block_base  <=  Qtr_block_base_map(block_row_r, block_col_r);

    a_addr_o    <=  E_BASE_ADDR + block_col_r*P + ((b_i + rel_row_r) mod P);
    b_addr_o    <=  QTR_BASE_ADDR + Qtr_block_base + i_r;
    z_o         <=  (a_i + 1) mod 2;
    z_addr_o    <=  E_BASE_ADDR + block_col_r*P + ((b_i + rel_row_r) mod P);

    seq: process(clk_i, rst_n_i)
    begin
        if rst_n_i = '0' then
            state       <=  IDLE;
            cur_idx_r   <=  0;
            block_row_r <=  0;
            block_col_r <=  0;
            rel_row_r   <=  0;
            i_r         <=  0;
        elsif rising_edge(clk_i) then
            state       <=  next_state;
            cur_idx_r   <=  cur_idx_c;
            block_row_r <=  block_row_c;
            block_col_r <=  block_col_c;
            rel_row_r   <=  rel_row_c;
            i_r         <=  i_c;
        end if;
    end process seq;
    comb: process(
        state, start_i,
        block_row_r, block_col_r, rel_row_r,
        row_idx_i, cur_idx_r, i_r,
        Qtr_block_count)
    begin
        next_state  <=  state;
        cur_idx_c   <=  cur_idx_r;
        block_row_c <=  block_row_r;
        block_col_c <=  block_col_r;
        rel_row_c   <=  rel_row_r;
        i_c         <=  0;
        wr_o        <=  '0';
        done_o      <=  '0';
        case (state) is
            when IDLE =>
                cur_idx_c   <=  0;
                block_row_c <=  0;
                block_col_c <=  0;
                if start_i = '1' then
                    next_state  <=  RELATIVE_ROW;
                end if;
            when RELATIVE_ROW =>
                if row_idx_i(cur_idx_r) /= INVALID then
                    block_row_c <=  row_idx_i(cur_idx_r) / P;
                    rel_row_c   <=  row_idx_i(cur_idx_r) mod P;
                    next_state  <=  SUM_ROW;
                else
                    next_state  <=  DONE;
                end if;
            when SUM_ROW =>
                wr_o        <=  '1';
                if i_r /= Qtr_block_count then
                    i_c         <=  i_r + 1;
                else
                    if block_col_r /= N0-1 then
                        block_col_c <=  block_col_r + 1;
                    else
                        if cur_idx_r = 15 then
                            next_state  <=  DONE;
                        else
                            cur_idx_c   <=  cur_idx_r + 1;
                            block_col_c <=  0;
                            next_state  <=  RELATIVE_ROW;
                        end if;
                    end if;
                end if; 
            when DONE =>
                done_o      <=  '1';
                if start_i = '0' then
                    next_state  <=  IDLE;
                end if;
        end case;
    end process comb;

end architecture rtl;

