import argparse
import matrix_generation

parser = argparse.ArgumentParser()
parser.add_argument("-n0", type=int, default=2, help="number of circulant blocks in H")
parser.add_argument("-p", type=int, default=155, help="size of circulant blocks")
parser.add_argument("-dv", type=int, default=17, help="weight of any block of H")
parser.add_argument("-m", type=int, nargs='+', default=[4, 3], help="weight of blocks of Q, check documentation")
parser.add_argument("--run_dense_comparison", action="store_true", help="use numpy libraries to check results")
options = parser.parse_args()

# Renaming for brevity
n0 = options.n0
p = options.p
dv = options.dv
m = options.m

# Check input sanity
if dv > p:
    print("dv should be smaller than p")
if len(m) != n0:
    print("m should contain n0 elements")
for i in range(n0):
    if m[i] > p:
        print("all values of m should be smaller than p")

if __name__ == '__main__':
    # Get full weight matrix for Qtr
    m_copy = list(m)
    m_copy.reverse()
    mtr_exp = []
    for i in range(n0):
        mtr_exp.append(m_copy[-i - 1:] + m_copy[:-i - 1])

    # Find limits to Qtr blocks
    limits = [[0 for x in range(n0)] for y in range(n0)]
    for i in range(n0):
        for j in range(n0):
            limits[i][j] = i*sum(m) + sum(mtr_exp[i][:j+1])

    # Generate matrices as we will have in memory
    Htr = matrix_generation.generate_Htr(n0, p, dv)
    Qtr = matrix_generation.generate_Qtr(n0, p, m)  # use a copy to make sure m is not modified
    # Save matrices in binary and decimal format
    with open("Htr.txt", 'w') as fd:
        for pos in Htr:
            fd.write(format(pos, "016b")+'\n')
    with open("Htr_dec.txt", "w") as fd:
        row = 0
        cnt = 0
        for pos in Htr:
            fd.write("{:d} ".format(pos))
            cnt += 1
            if cnt == dv:
                fd.write('\n')
                cnt = 0
                row += 1
    with open("Qtr.txt", 'w') as fd:
        for pos in Qtr:
            fd.write(format(pos, "016b") + '\n')
    with open("Qtr_dec.txt", "w") as fd:
        row = 0
        col = 0
        cnt = 0
        for pos in Qtr:
            fd.write("{:d} ".format(pos))
            cnt += 1
            if cnt == mtr_exp[row][col]:
                cnt = 0
                if col != n0-1:
                    fd.write(10 * ' ')
                    col += 1
                else:
                    fd.write('\n')
                    col = 0
                    row += 1

    # Initialize Ltr
    Ltr = [p for x in range(n0*sum(m)*dv)]

    # Get Ltr = Qtr*Htr
    Qtr_index = 0  # index in Qtr
    for Qtr_block_i in range(n0):  # rows of blocks of Qtr
        for Qtr_block_j in range(n0):  # columns of blocks of Qtr
            Htr_block_i = Qtr_block_j  # corresponding to rows of 1 block of Htr
            aux_block = [p for x in range(mtr_exp[Qtr_block_i][Qtr_block_j] * dv)]
            aux_index = 0  # target index in aux_block

            # Get the tentative result for this block combination
            while Qtr_index < limits[Qtr_block_i][Qtr_block_j]:
                Htr_index = dv*Htr_block_i  # warp back to appropriate value
                while Htr_index < dv*(Htr_block_i + 1):
                    # print("{:<10s}{:3d}{:5s}{:<10s}{:3d}{:5s}{:<10s}{:3d}{:5s}{:<10s}{:3d}{:5s}{:<10s}{:3d}".format(
                    #     "i:", Qtr_block_i, "", "j:", Qtr_block_j, "", "Qtr:", Qtr_index, "", "Htr:", Htr_index, "",
                    #     "aux:", aux_index
                    # ))
                    aux_block[aux_index] = (Qtr[Qtr_index] + Htr[Htr_index]) % p
                    Htr_index += 1
                    aux_index += 1
                Qtr_index += 1

            # Sort the tentative result
            aux_block_builtin = sorted(aux_block)
            i = 1
            while i < len(aux_block):
                tmp = aux_block[i]
                j = i-1
                while j >= 0 and aux_block[j] > tmp:
                    aux_block[j+1] = aux_block[j]
                    j -= 1
                aux_block[j+1] = tmp
                i += 1
            # closer to hardware, used to test it
            # i = 1
            # while i < len(aux_block):
            #     tmp = aux_block[i]
            #     j = i - 1
            #     while j != 0:
            #         if aux_block[j] <= tmp:
            #             break
            #         aux_block[j + 1] = aux_block[j]
            #         j -= 1
            #     if j == 0 and aux_block[j] > tmp:
            #         aux_block[j + 1] = aux_block[j]
            #         aux_block[j] = tmp
            #     else:
            #         aux_block[j + 1] = tmp
            #     i += 1
            assert aux_block == aux_block_builtin

            # Simplify
            i = 0
            j = 0
            while i < len(aux_block)-1:
                if aux_block[i] == aux_block[i+1]:
                    i += 1
                else:
                    aux_block[j] = (aux_block[i])
                    j += 1
                i += 1
            if i == len(aux_block) - 1:
                aux_block[j] = aux_block[i]
                j += 1
            # Fill the rest with invalid values
            while j < len(aux_block):
                aux_block[j] = p
                j += 1

            # Sum and simplify with Ltr
            i = 0
            j = Qtr_block_i*sum(m)*dv
            k = 0
            tmp = [p for x in range(sum(m)*dv)]
            while i < len(aux_block) and j < (Qtr_block_i+1)*sum(m)*dv and aux_block[i] != p and Ltr[j] != p:
                # If we encounter two values which are the same we do not copy them
                if aux_block[i] == Ltr[j]:
                    i += 1
                    j += 1
                elif aux_block[i] < Ltr[j]:
                    tmp[k] = aux_block[i]
                    i += 1
                    k += 1
                else:
                    tmp[k] = Ltr[j]
                    j += 1
                    k += 1
            while i < len(aux_block) and aux_block[i] != p:
                tmp[k] = aux_block[i]
                i += 1
                k += 1
            while j < limits[Qtr_block_i][Qtr_block_j]*dv and Ltr[j] != p:
                tmp[k] = Ltr[j]
                j += 1
                k += 1
            while k < sum(m)*dv:
                tmp[k] = p
                k += 1

            # copy result in Ltr
            Ltr[Qtr_block_i*sum(m)*dv:(Qtr_block_i+1)*sum(m)*dv] = list(tmp)

    if options.run_dense_comparison:
        print("Checking equivalence with dense representation...")
        # Check
        # Rebuild matrices
        Htr_dense = matrix_generation.Htr_sparse_to_dense(Htr, n0, p, dv)
        Qtr_dense = matrix_generation.Qtr_sparse_to_dense(Qtr, n0, p, m)
        Ltr_dense = Qtr_dense.dot(Htr_dense) % 2
        assert (Ltr_dense == matrix_generation.Ltr_sparse_to_dense(Ltr, n0, p, dv, m)).all()

    with open("Ltr.txt", 'w') as fd:
        for pos in Ltr:
            fd.write(format(pos, "016b")+'\n')
    with open("Ltr_dec.txt", "w") as fd:
        row = 0
        cnt = 0
        for pos in Ltr:
            fd.write("{:d} ".format(pos))
            cnt += 1
            if cnt == sum(m)*dv:
                fd.write('\n')
                cnt = 0
                row += 1

